'''
Function to read data in MFC BIRA-IASB binary format supported by QDOAS

Author:  Caroline FAYT (Royal Belgian Institute for Space Aeronomy)
Date:  February 2018
Version:
'''

import struct;

from numpy import *;
from doas_constants import *
from doas_lib import *

class MFC_BIRA_spectrum :
    """
    Definition of a spectrum in MFC BIRA-IASB binary format

         header            : general information on the record (see complete description below)                                         \n
         spectrum          : spectrum                                                                                                   \n
         spectrumCorrected : if offsets and dark currents are measured and saved in the spectra file, spectra are corrected             \n

    Description of the header
    -------------------------

         ==  ====================  ===========  ======  =====================================================================
         no  name                  type         length  comment
         ==  ====================  ===========  ======  =====================================================================
          1  'measurementType'     'int32'          1   the measurement type
          2  'da_year'             'int16'          1   year part of the measurement date
          3  'da_day'              'int8'           1   day part of the measurement date
          4  'da_month'            'int8'           1   month part of the measurement date
          5  'start_ti_min'        'int8'           1   min part of the starting measurement time
          6  'start_ti_hour'       'int8'           1   hour part of the starting measurement time
          7  'start_ti_hund'       'int8'           1   hundredth of sec part of the starting measurement time
          8  'start_ti_sec'        'int8'           1   secunds part of the starting measurement time
          9  'end_ti_min'          'int8'           1   min part of the ending measurement time
         10  'end_ti_hour'         'int8'           1   hour part of the ending measurement time
         11  'end_ti_hund'         'int8'           1   hundredth of sec part of the ending measurement time
         12  'end_ti_sec'          'int8'           1   secunds part of the ending measurement time
         13  'spectroName'         'int8'          64   the name of the spectrometer
         14  'deviceName'          'int8'          64   the name of the device
         15  'scansNumber'         'int32'          1   the number of scans
         16  'exposureTime'        'real*4'         1   the exposure time
         17  'totalAcqTime'        'real*4'         1   the total exposure time (scansNumber x exposureTime)
         18  'siteName'            'int8'          64   the name of the observation site
         19  'longitude'           'real*4'         1   the longitude of the observation site
         20  'latitude'            'real*4'         1   the latitude of the observation site
         21  'azimuthAngle'        'real*4'         1   the viewing azimuth angle
         22  'elevationAngle'      'real*4'         1   the viewing elevation angle
         23  'fileName'            'int8'        1024   the original file name
         24  'temperature'         'real*4'         1   the temperature
         ==  ====================  ===========  ======  =====================================================================

    Equivalent of the header in Python
    ----------------------------------

         '=ih10B64s64si2f64s4f1024sf'

    with :

         =========  ======================  ========  ==========
         Character  Byte order              Size      Alignment
         =========  ======================  ========  ==========
         @          native                  native    native
         =          native                  standard  none
         <          little-endian           standard  none
         >          big-endian              standard  none
         !          network (= big-endian)  standard  none
         =========  ======================  ========  ==========

         ======  ==============  ==================  ======================
         Format  C Type          Python type         Standard size  (byte)}     \n
         ======  ==============  ==================  ======================
         x       pad             byte                no value                   \n
         c       char            string of length 1  1                          \n
         b       signed char     integer             1                          \n
         B       unsigned char   integer             1                          \n
         ?       _Bool           bool                1                          \n
         h       short           integer             2                          \n
         H       unsigned short  integer             2                          \n
         i       int             integer             4                          \n
         I       unsigned int    integer             4                          \n
         l       long            integer             4                          \n
         L       unsigned long   integer             4                          \n
         q       long long       integer             8                          \n
         Q       unsigned long   long  integer       8                          \n
         f       float           float               4                          \n
         d       double          float               8                          \n
         s       char[]          string                                         \n
         p       char[]          string                                         \n
         ======  ==============  ==================  ======================
    """

    def __init__(self,*args):
       '''
       constructor for MFC_BIRA_spectrum: initialize the class variables and arrays

       '''

       self.header=args[0];                                                      # header
       self.spectrum=args[1];                                                    # original spectrum
       self.spectrumCorrected=args[1];                                           # spectrum dark current and offset corrected (operation to make later)

MFC_BIRA_fields= ('measurementType',   \
                  'da_year',           \
                  'da_day',            \
                  'da_month',          \
                  'start_ti_min',      \
                  'start_ti_hour',     \
                  'start_ti_hund',     \
                  'start_ti_sec',      \
                  'end_ti_min',        \
                  'end_ti_hour',       \
                  'end_ti_hund',       \
                  'end_ti_sec',        \
                  'spectroName',       \
                  'deviceName',        \
                  'scansNumber',       \
                  'exposureTime',      \
                  'totalAcqTime',      \
                  'siteName',          \
                  'longitude',         \
                  'latitude',          \
                  'azimuthAngle',      \
                  'elevationAngle',    \
                  'fileName',          \
                  'temperature');

MFC_BIRA_headerFmt = '=ih10B64s64si2f64s4f1024sf'
MFC_BIRA_headerLen = struct.calcsize(MFC_BIRA_headerFmt)
MFC_BIRA_headerUnpack = struct.Struct(MFC_BIRA_headerFmt).unpack_from

def MFC_BIRA_ReadNextRecord(f,spectraSize):
    """
    return the next MFC record whatever the measurement type (observation, offset, dark current)

    Parameters
    ----------

    f : pointer to the file to read

    Returns
    -------

    MFC_BIRA_record : the next record

    See also
    --------

    MFC_BIRA_spectrum,MFC_BIRA_ReadSpe
    """

    # Read header

    bytesPacked = f.read(MFC_BIRA_headerLen)                                         # read MFC_BIRA_headerLen bytes
    if not bytesPacked: return []
    bytesUnpacked = MFC_BIRA_headerUnpack(bytesPacked)                               # decomposition of the bytes pack according to the format string

    # Associate the list of header fields with the list of items in s

    header=dict(zip(MFC_BIRA_fields, bytesUnpacked))                                 # map the two lists into a dictionary
    spectrum=array(reshape(fromfile(f,dtype=float32,count=int(spectraSize)),(int(spectraSize),),order='F'),dtype=float32);

    # Save record

    MFC_BIRA_record=MFC_BIRA_spectrum(header,spectrum);

    # Return

    return MFC_BIRA_record;

def MFC_BIRA_ReadSpe(speFile,correctionFlag,averageFlag):
    """
    Read spectra measured by MFC detectors

    Parameters
    ----------

    speFile : the name of the spectra file to load                                                        \n
    correctionFlag : 1 to correct spectra by the dark current and offset if they are present in the file  \n
                     0 otherwise                                                                          \n
    averageFlag : 1 to average spectra with the number of scans                                           \n

    Returns
    -------

    MFC_BIRA_spe : the spectra retrieved from the file

    See also
    --------

    MFC_BIRA_ReadNextRecord
    """

    MFC_BIRA_spe=[];                                                     # array of spectra to return
    drkTint=0.;
    noff=0;
    ndrk=0;

    functionName="MFC_BIRA_ReadSpe"

    if not os.path.isfile(speFile):
        print('['+functionName+'] '+speFile+' not found !!!')
    else:

        f=open(speFile,"rb")

        file_header=f.read(8);

        if (not file_header):
          print('['+functionName+'] '+speFile+' empty or wrong format !!!')
        else:


           (nrecords,npixels)=struct.Struct('=2i').unpack_from(file_header);

           off=zeros((npixels,));
           drk=zeros((npixels,));

           while True:

               # Read next record

               spe=MFC_BIRA_ReadNextRecord(f,npixels)
               if size(spe)==0 : break;

               # Read any spectrum (except dark current and offset if correction is requested)

               if not correctionFlag or ((spe.header['measurementType']!=QDOAS_MeasType_dark) and (spe.header['measurementType']!=QDOAS_MeasType_offset) and (spe.header['measurementType']!=QDOAS_MeasType_lamp)):
                  MFC_BIRA_spe.append(spe);

               # If dark current and offset correction is requested, keep these spectra apart

               elif (spe.header['measurementType']==QDOAS_MeasType_dark):
                 drk=drk+spe.spectrum;
                 drkTint=spe.header['exposureTime'];
                 ndrk=ndrk+spe.header['scansNumber'];
               elif (spe.header['measurementType']==QDOAS_MeasType_offset):
                 off=off+spe.spectrum;
                 noff=noff+spe.header['scansNumber'];

           if correctionFlag :

             if (noff):
                off=off/noff;                       # Average offset (account for the number of spectra and the number of scans)

             if (ndrk and drkTint!=0.):
                drk=(drk-off*ndrk)/(drkTint*ndrk);       # Average dark current and correct by the offset

             for spe in MFC_BIRA_spe:

                spe.spectrumCorrected=(spe.spectrum-off*spe.header['scansNumber']-drk*spe.header['scansNumber']*spe.header['exposureTime'])
                if averageFlag:
                   spe.spectrumCorrected=spe.spectrumCorrected/spe.header['scansNumber'];
           else:
           
             for spe in MFC_BIRA_spe:
               if averageFlag: 
                  spe.spectrum=spe.spectrum/spe.header['scansNumber'];
               spe.spectrumCorrected=spe.spectrum;


        f.close();

    # Return

    return MFC_BIRA_spe;
   
def MFC_BIRA_SaveRecord(g,spe):

    """
    MFC_BIRA_SaveRecord(g,spe)

    Save a spectrum record in the output file

    Parameters
    ----------

    g   : pointer to the output file
    spe : spectrum record to save (MFC_BIRA_spectrum structure)

    See also
    --------

    MFC_BIRA_spectrum
    """

    header=spe.header;
    spectrum=array(spe.spectrumCorrected,dtype=float32);

    bytesPacked=struct.pack(MFC_BIRA_headerFmt, \
                            header['measurementType'], \
                            header['da_year'        ], \
                            header['da_day'         ], \
                            header['da_month'       ], \
                            header['start_ti_min'   ], \
                            header['start_ti_hour'  ], \
                            header['start_ti_hund'  ], \
                            header['start_ti_sec'   ], \
                            header['end_ti_min'     ], \
                            header['end_ti_hour'    ], \
                            header['end_ti_hund'    ], \
                            header['end_ti_sec'     ], \
                            header['spectroName'    ], \
                            header['deviceName'     ], \
                            header['scansNumber'    ], \
                            header['exposureTime'   ], \
                            header['totalAcqTime'   ], \
                            header['siteName'       ], \
                            header['longitude'      ], \
                            header['latitude'       ], \
                            header['azimuthAngle'   ], \
                            header['elevationAngle' ], \
                            header['fileName'       ], \
                            header['temperature'    ]);
    
    g.write(bytesPacked);
    g.write(spectrum);

def MFC_BIRA_SaveFile(fileName,spe):

    """
    MFC_BIRA_SaveFile(fileName,spe)

    Save a list of spectrum records in the output file

    Parameters
    ----------

    fileName : the name of the output file
    fileName : the list of spectrum records to save

    See also
    --------

    MFC_BIRA_spectrum, MFC_BIRA_SaveRecord
    """

    if len(spe):

       try :
          g=open(fileName,"wb");

       except IOError :

          print(sprintf("Can not open file %s in writing mode\n",fileName));

       else :

          g.write(struct.pack('2i',len(spe),len(spe[0].spectrum)));

          for i in arange(0,len(spe)):
             MFC_BIRA_SaveRecord(g,spe[i]);

          g.close();

def MFC_BIRA_Calib(calibFile,speFile,newCalibFile,newSpeFile):

    """
    MFC_BIRA_spe=MFC_BIRA_Calib(calibFile,speFile,newCalibFile,newSpeFile)

    If a file doesn't contain calibration measurements, complete it with old ones
    Otherwise, save these calibration measurements to update another spectra file

    Parameters
    ----------

    calibFile    : the name of the file with calibration measurements           \n
    speFile      : the name of the file (eventually without calibration measurements)\n
    newCalibFile : the name of the file with new calibration measurements
    newSpeFile   : the name of the (speFile with offset and dark current from calibFile) \n

    See also
    --------

    MFC_BIRA_ReadSpe,MFC_BIRA_SaveFile
    """

    if not os.path.isfile(calibFile):
       print('[MFC_BIRA_Calib] '+calibFile+' not found !!!');
       calibFromClb=[];
    else:
       clb=MFC_BIRA_ReadSpe(calibFile,0,0);
       calibFromClb=[s for (i,s) in enumerate(clb) if ((s.header['measurementType']==QDOAS_MeasType_dark) or (s.header['measurementType']==QDOAS_MeasType_offset))];
       
       
    if not os.path.isfile(speFile):
       print('[MFC_BIRA_Calib] '+speFile+' not found !!!');
    else:

       # Read input files

       spe=MFC_BIRA_ReadSpe(speFile,0,0);

       # Get offset and dark current records

       calibFromSpe=[s for (i,s) in enumerate(spe) if ((s.header['measurementType']==QDOAS_MeasType_dark) or (s.header['measurementType']==QDOAS_MeasType_offset))];

       print(sprintf("Number of calibration measurements : calib %d spe %d\n",len(calibFromClb),len(calibFromSpe)));

       # The spectra files contains calibration measurements

       if (len(calibFromSpe)>0):

          calibFromSpe=[s for (i,s) in enumerate(spe) if ((s.header['measurementType']==QDOAS_MeasType_dark) or (s.header['measurementType']==QDOAS_MeasType_offset) or (s.header['measurementType']==QDOAS_MeasType_lamp))];

          print(sprintf("New calibration measurements saved in %s\n",newCalibFile));

          MFC_BIRA_SaveFile(newCalibFile,calibFromSpe);                          # Save new calibration measurments
          MFC_BIRA_SaveFile(newSpeFile,spe);                                     # Save original spectra in the new spectra file

       # if the spectra file doesn't contain calibration records, complete it with dark current and offset measurements from the input calibration file

       elif (len(calibFromClb)>0) :

          for s in spe :
             calibFromClb.append(s);                                          # append spectra to calibration measurements

          print(sprintf("Update %s with calibration measurements from %s\n",newSpeFile,calibFile));

          MFC_BIRA_SaveFile(newSpeFile,calibFromClb);

       # the calibration file doesn't contain calibration measurements -> error

       else:

          print(sprintf("Calibration file %s do not contain calibration measurements\n",calibFile));
