'''
Tools to handle XML files


    The purpose of this module is to load xml file in a structure whose fields
    are the tags and attributes.  The levels of tags are respected.

    Given the following xml file example.xml

       <root>
          <tag1 attr1="...">
             <tag2 attr2="..." />
             <tag3 attr3="..." /"
          </tag1>
       </root>

    a=XML_LoadStructure("example.xml") returns a structure a as follow :

       a.tag1.attr1
       a.tag1.tag2.attr2
       a.tag1.tag3.attr3

    If a same tag is repeated, a structure array is created :

       <root>
          <tag1 attr1="...">
             <tag2 attr2="..." />
             <tag2 attr3="..." /"
          </tag1>
       </root>

    will return a structure a as follows :

       a.tag1.attr1
       a.tag1.tag2[0].attr2
       a.tag1.tag2[1].attr3

Author:  Caroline FAYT (Royal Belgian Institute for Space Aeronomy)
Date: 26/09/2017
History : 14/04/2021 use list(e) instead of getchildren with Python 3.9
Version:
'''


# ------------------------------------------------------------------------------
# MODULE        XML_tools
# ------------------------------------------------------------------------------
# PURPOSE       Tools to handle XML files
# ------------------------------------------------------------------------------
# AUTHOR        Caroline FAYT
# COPYRIGHT     Royal Belgian Institute for Space Aeronomy
# CREATION DATE 26/09/2017
#
# MODIFIED
#
# ------------------------------------------------------------------------------

from doas_lib import *

import xml.etree.cElementTree as et
import csv

# =======================
# DEFINITION OF THE CLASS
# =======================

# the structure can be used as a dictionary (useful for JSON)

class XML_structure(dict) :

   # Create the structure

   def __init__(self):

      self.__dict__ = self
      pass;

   # Returns tags and values pairs at a given level

   def GetFields(self):

     tags=[];
     vals=[];
     for attribute, value in self.__dict__.items():
       tags.append(attribute)
       vals.append(value);
     return(tags,vals)

# ==============
# LOAD XML FILES
# ==============

#
# XML_AddElements : add recursively and XML e (and its children) in the structure
#

def XML_AddElements(xmlStruct,e):

    lastTag='';
    childList=[];
    n=0;

    # Add attributes first

    for i in e.attrib:
       setattr(xmlStruct,i,e.attrib[i]);

    # Add children tags

    for i in list(e): # e.getchildren() for python versions < 3.9

       b=XML_structure();
       XML_AddElements(b,i);

       # If there are similar tags at the same level, create a list
       # In the test below, the last sequence of similar tags is close

       if lastTag!='' and i.tag!=lastTag and n>0 :

          setattr(xmlStruct,lastTag,childList[0] if n==1 else childList);
          childList=[];
          n=0;

       # Build the list of tags

       childList.append(b);
       lastTag=i.tag;
       n=n+1;

    # Flush last tags

    if n>0:
       setattr(xmlStruct,lastTag,childList[0] if n==1 else childList);

#
# XML_LoadStructure : first function to call (see comments above)
#

def XML_LoadStructure(configFile):

    """
    XML_LoadStructure(configFile)

    Load the content of a XML file in a structure

    Parameters
    ----------

    configFile : the xml file to read                                         \n

    Returns
    -------
    a structure whose fields are retrieved from the different tags and attributes of
    the xml file
    """

    xmlStruct=XML_structure();                                                  # create a empty structure

    if not os.path.isfile(configFile):
       print('[XML_LoadStructure] '+configFile+' not found !!!');
    else:

       f=et.parse(configFile);                                                  # parse the xml tree
       root=f.getroot();                                                        # get the root

       XML_AddElements(xmlStruct,root);                                         # add recursively attributes

    return xmlStruct;                                                           # return

def XML_GetPath(root,path):

    fields=path.split('/');
    xmlStruct=root;
    rc=0;

    for str in fields:
       if str != '' :
          xmlFields=xmlStruct.GetFields();
          if str in xmlFields[0]:
             xmlStruct=getattr(xmlStruct,str);
          else:
             rc=1;
             xmlStruct=[];
             break;

    if (xmlStruct==[]):
       xmlStruct='';

    return xmlStruct;

def listToStringWithoutBrackets(list1):
    return str(list1).replace('[','').replace(']','');

def XML_GetAttributes(configFile,paramList):

    a=[];

    if not os.path.isfile(configFile):
       print('[XML_LoadStructure] '+configFile+' not found !!!');
    else:

       xmlStruct=XML_LoadStructure(configFile)

       for param in paramList:
           a.append(XML_GetPath(xmlStruct,param))

       print(listToStringWithoutBrackets(a))

# ==========================================
# COMPARISON OF THE CONTENT OF XML STRUCTURE
# ==========================================

# XML_IsType : check that all elements of structList have the same expected type

def XML_IsType(structList,expectedType):

    rc=True;

    for i in structList:
       if not isinstance(i,expectedType):
          rc=False;
          break;

    return rc;

# XML_IsSize : check that all elements of structList have the same size

def XML_IsSize(structList,expectedSize):

    rc=True;

    for i in structList:
       if len(i)!=expectedSize:
          rc=False;
          break;

    return rc;

# XML_GetListIndex : when elements of structList are also lists, return a list
# composed with all the items at the given position

def XML_GetListIndex(structList,position):

    fieldsList=[];

    for i in structList:
       fieldsList.append(i[position]);

    return fieldsList;

# XML_GetValue : when elements of structList are simple tag/value pairs, return
# the list of value

def XML_GetValue(structList,tag):

    fieldsList=[];

    for i in structList:
       (tags,value)=i.GetFields()
       val=[v for (t,v) in zip(tags,value) if (t==tag)];
       if (len(val)):
        fieldsList.append(val[0]);
       else:
        fieldsList.append([]);

    return fieldsList;

# XML_CompareValues : compare the elements of structList with the given value

def XML_CompareValues(structList,value):

    rc=True;

    for i in structList:
       if i!=value:
          rc=False
          break;

    return rc;

def XML_Compare(structList,*args):

#    The first function to call is XML_LoadStructure.  Then structures could be
#    compare with the function XML_Compare.  Comparisons can be performed
#    between two or more elements and from root or from a lower level.
#
#    The main application at BIRA-IASB could be to compare QDOAS XML files
#
#    for example, q1=XML_LoadStructure("qdoas_config_1.xml")
#                 q2=XML_LoadStructure("qdoas_config_2.xml")
#
#    We could compare the content of the two files :
#
#      XML_Compare([q1,q2]);
#
#   Or just the project :
#
#      XML_Compare([q1.project,q2.project]);
#
#   Or the analysis windows :
#
#      XML_Compare([q1.project.analysis_windows,q2.project.analysis_windows]);
#
#   Note that if there are several project or analysis_windows, and only one item
#   has to be compared, we could have something like that :
#
#      XML_Compare([q1.project[0].analysis_windows[1],q2.project[0].analysis_windows[1]]);
#
#   The analysis windows of a same project of the same file can be compared
#
#      XML_Compare([q1.project[0].analysis_windows[1],q1.project[0].analysis_windows[2]]);
#
#   Note that if there is no additional argument (*args), the root will be tagged 'root'

    rc=1;

    errMsg=''
    somethingToPrint=0
    currentTag='root' if len(args)==0 else args[0]
    deepFlag=1 if len(args)<2 else args[1]

    # If the input list doesn't contain at least two elements, it can not work

    if not isinstance(structList,list) or len(structList)<=1:
       print('[XML_Compare] Expect a list of at least 2 elements')

    # The items to compare should be of the same type

    elif not XML_IsType(structList,type(structList[0])):
       print('[XML_Compare] '+currentTag+' items do not have all the same type')

    # If the items are list themselves, they should have the same size

    elif isinstance(structList[0],list) and len(structList[0])>0 and XML_IsSize(structList,len(structList[0])):
       for i in arange(0,len(structList[0])):
          newTag=sprintf('%s[%d]',currentTag,i)
          XML_Compare(XML_GetListIndex(structList,i),newTag);

    # If the items of the list are also XML_structure, recursive comparison

    elif isinstance(structList[0],XML_structure) :
       (tags,value)=structList[0].GetFields();
       for (tag,val) in zip(tags,value):
           newTag=sprintf('%s.%s',currentTag,tag);
           XML_Compare(XML_GetValue(structList,tag),newTag);

    # End of the chain : we compare attributes/values pairs

    elif not XML_CompareValues(structList,structList[0]):
      print(sprintf('%s\n',currentTag))
      for i,item in enumerate(structList):
          print(sprintf('   %2d: %s',i,item));
      print
      # errMsg='tag is missing or has different values';
      somethingToPrint=1
    else:
      somethingToPrint=1

    # if (somethingToPrint):
    #
    #     filename=r'W:\projects\FRM4DOAS\programs\trunk\python\tools\toto.dat';
    #     a=[]
    #     a.append(currentTag)
    #     for i in structList:
    #        a.append(i)
    #     a.append(errMsg)
    #
    #     if os.path.exists(filename):
    #         append_write = 'a' # append if already exists
    #     else:
    #         append_write = 'w' # make a new file if not
    #
    #
    #
    #     with open(filename, append_write) as f:
    #         writer = csv.writer(f)
    #         writer.writerow(a)

