'''
Collection of utility functions.  Most of them have an equivalent in Matlab.

Author:  Caroline FAYT (Royal Belgian Institute for Space Aeronomy)
Date:  2013 -- 2019
Version:
'''

#! /usr/bin/env python3

import os
import warnings
import pkgutil;

pandasFlag=1 if pkgutil.find_loader("pandas") else 0

if pandasFlag:
   from pandas import * # - not supported on kronos

from doas_constants import *
from pylab import *
from datetime import *
from scipy import interpolate
from scipy.interpolate import interp1d
from scipy.interpolate import interp2d
from scipy.linalg.blas import dgemm
from scipy.signal import lfilter
# NEW IN 2.7 from math import erf as erf
import numpy as np
import subprocess as sp
import re

#
# HISTORY
#
# 2017-02-24 : interp1d
# 2017-03-08 :
#
#     New function : MJD2000
#     dec2dms,sunangl,sunangl_date : apply on scalars and vectors (before, scalars only)
#
# 2019-02-25 :
#
#     loctime (add correction of dates for high longitudes)

# ------------------------
# OTHER IMPORTED LIBRARIES
# ------------------------

#  Copyright (c) 2013 Alexandre Drouin. All rights reserved.
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of
#  this software and associated documentation files (the "Software"), to deal in
#  the Software without restriction, including without limitation the rights to
#  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
#  of the Software, and to permit persons to whom the Software is furnished to do
#  so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  If you happen to meet one of the copyright holders in a bar you are obligated
#  to buy them one pint of beer.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.

def linear_least_squares(a, b, residuals=False):
    """
    Return the least-squares solution to a linear matrix equation.

    Solves the equation `a x = b` by computing a vector `x` that
    minimizes the Euclidean 2-norm `|| b - a x ||^2`.  The equation may
    be under-, well-, or over- determined (i.e., the number of
    linearly independent rows of `a` can be less than, equal to, or
    greater than its number of linearly independent columns).  If `a`
    is square and of full rank, then `x` (but for round-off error) is
    the "exact" solution of the equation.

    Parameters
    ----------
    a : (M, N) array_like
        "Coefficient" matrix.
    b : (M,) array_like
        Ordinate or "dependent variable" values.
    residuals : bool
        Compute the residuals associated with the least-squares solution

    Returns
    -------
    x : (M,) ndarray
        Least-squares solution. The shape of `x` depends on the shape of
        `b`.
    residuals : int (Optional)
        Sums of residuals; squared Euclidean 2-norm for each column in
        ``b - a*x``.
    """
    if type(a) != np.ndarray or not a.flags['C_CONTIGUOUS']:
        warnings.warn('Matrix a is not a C-contiguous numpy array. The solver will create a copy, which will result' + \
             ' in increased memory usage.')

    a = np.asarray(a, order='c')
    i = dgemm(alpha=1.0, a=a.T, b=a.T, trans_b=True)
    x = np.linalg.solve(i, dgemm(alpha=1.0, a=a.T, b=b)).flatten()

    if residuals:
        return x, np.linalg.norm(np.dot(a, x) - b)
    else:
        return x


if __name__ == "__main__":
    x = np.array([0, 1, 2, 3])
    y = np.array([-1, 0.2, 0.9, 2.1])

    A = np.vstack([x, np.ones(len(x))]).T
    A = np.asarray(A, order='c')

    m, c = linear_least_squares(A, y)

    print (m, c)

    import matplotlib.pyplot as plt

    plt.plot(x, y, 'o', label='Original data', markersize=10)
    plt.plot(x, m * x + c, 'r', label='Fitted line')
    plt.legend()
    plt.show()

# --------------
# MATH FUNCTIONS
# --------------

def linfit(x,y,dy):
    """
    Calculates a weighted linear least-square adjustment on a couple of         \n
    vectors x, y, with dy the estimated errors on y.                            \n

    Returns
    -------

       p = linfit(X,Y,dY)     with  Y=p(2)+p(1)*X

    See also
    --------

    linval
    """

    dy2=dy*dy
    X  = sum ( x/dy2 );
    X2 = sum ( x*x/dy2 );
    Y  = sum ( y/dy2 );
    XY = sum ( x*y/dy2 );
    O  = sum ( 1./dy2 );

    Delta = O*X2 - X*X;

    a = ( X2*Y - X*XY ) / Delta;
    b = ( O*XY - X*Y )  / Delta;

    p=array([b,a]);

    return(p);

def linval(x,y,xp):
    """
    Linear interpolation

    Parameters
    ----------

    x   original abscissas
    y   original ordinates
    xp  new abscissas

    Returns
    -------
    YP new ordinates calculated in new xp abscissas

    See also
    --------

    linfit
    """

    np=len(xp);
    nn=len(x);

    if np==0 or nn==0:
       yp=array([]);
    else:
       yp=zeros(len(xp));

       for i in arange(0,np):

          if (xp[i]>=x[0] and xp[i]<=x[nn-1]):
             for k in arange(0,nn-1):
                if (x[k]<=xp[i] and xp[i]<=x[k+1]):
                   break;

          a=(y[k+1]-y[k])/(x[k+1]-x[k]);
          b=y[k]-a*x[k];
          yp[i]=a*xp[i]+b;

    return yp;

def polyfit2(x,y,e,n):
    """
    Translation of the polyfit2 Matlab function : weighted fit polynomial to data

    POLYFIT2(X,Y,E,N) finds the coefficients of a polynomial P(X) of
    degree N that fits the data, P(X(I))~=Y(I), in a least-squares sense.
    The fit is weighted by data errors E.

    [P,S] = POLYFIT2(X,Y,E,N) returns the polynomial coefficients P and a
    structure S for use with POLYVAL to obtain error estimates on
    predictions.  If the errors in the data, Y, are independent normal
    with constant variance, POLYVAL will produce error bounds which
    contain at least 50% of the predictions.

    The structure S contains the Cholesky factor of the Vandermonde
    matrix (R), the degrees of freedom (df), and the norm of the
    residuals (normr) as fields.

    See also POLY, POLYVAL, ROOTS.

    J.N. Little 4-21-85, 8-23-86; CBM, 12-27-91 BAJ, 5-7-93.
    Copyright (c) 1984-98 by The MathWorks, Inc.
    $Revision: 5.9 $  $Date: 1997/11/21 23:40:57 $
    $Modified by MVR: 2000/03/05

    The regression problem is formulated in matrix format as:

       y = V*p    or

             3  2
       y = [x  x  x  1] [p3
                         p2
                         p1
                         p0]

    where the vector p contains the coefficients to be found.  For a
    7th order polynomial, matrix V would be:

    V = [x.^7 x.^6 x.^5 x.^4 x.^3 x.^2 x ones(size(x))];

    Parameters
    ----------
                                                                                \n
    x,y  : the data                                                             \n
    e    : errors on the data                                                   \n
    n    : the degree of the polynomial

    Returns
    -------

    p : polynomial coefficients to build the polynomial that fits the data      \n
    s : (R,df,normr) with R     : the Cholesky factor of the Vandermonde matrix \n
                          df    : the degrees of freedom                        \n
                          normr : the norm of the residuals                     \n

    See also
    --------

    polyval2
    """

    p=array([]);
    R=array([]);
    df=0.;
    normr=0.;

    if (len(x)!=len(y)):
       print('[polyfit2] X and Y vectors must be the same size.');
    else:

       # Construct Vandermonde matrix

       V=vander(x,n+1);

       # Weight by errors

       V0=V;  V=(V.T/e).T;
       y0=y;  y=divide(y,e);

       # Solve least squares problem, and save the Cholesky factor.

       Q,R=qr(V,mode='full');                                                   # [Q,R] = qr(V,0);
       p=linear_least_squares(R,dot(Q.T,y))                                     # p = R\(Q'*y);    % Same as p = V\y;

       Q,R=qr(V0,mode='full');                                                   # [Q,R] = qr(V0,0);
       r = y0 - dot(V0,p);

       # s is a list containing three elements: the Cholesky factor of the
       # Vandermonde matrix, the degrees of freedom and the norm of the residuals.

       df=len(y)-(n+1);
       normr=norm(r);

    s=(R,df,normr);
    return p,s

def polyval2(p,x,*args):
    """
    Translation of the polyval Matlab function

    Y = POLYVAL(P,X) returns the value of a polynomial P evaluated at X. P
    is a vector of length N+1 whose elements are the coefficients of the
    polynomial in descending powers.

        Y = P(1)*X^N + P(2)*X^(N-1) + ... + P(N)*X + P(N+1)

    If X is a matrix or vector, the polynomial is evaluated at all
    points in X.  See also POLYVALM for evaluation in a matrix sense.

    [Y,DELTA] = POLYVAL(P,X,S) uses the optional output structure S created
    by POLYFIT to generate prediction error estimates DELTA.  DELTA is an
    estimate of the standard deviation of the error in predicting a future
    observation at X by P(X).

    If the coefficients in P are least squares estimates computed by
    POLYFIT, and the errors in the data input to POLYFIT are independent,
    normal, with constant variance, then Y +/- DELTA will contain at least
    50% of future observations at X.

    Y = POLYVAL(P,X,[],MU) or [Y,DELTA] = POLYVAL(P,X,S,MU) uses XHAT =
    (X-MU(1))/MU(2) in place of X. The centering and scaling parameters MU
    are optional output computed by POLYFIT.

    Class support for inputs P,X,S,MU:
       float: double, single

    See also POLYFIT, POLYVALM.

    Copyright 1984-2004 The MathWorks, Inc.
    $Revision: 5.16.4.4 $  $Date: 2004/07/05 17:01:38 $

    DELTA can be used to compute a 100(1-alpha)% prediction interval
    for future observations at X, as Y +/- DELTA*t(alpha/2,df), where
    t(alpha/2,df) is the upper (alpha/2) quantile of the Student's t
    distribution with df degrees of freedom.  Since t(.25,df) < 1 for any
    degrees of freedom, Y +/- DELTA is at least a 50% prediction interval
    in all cases.  For large degrees of freedom, the confidence level
    approaches approximately 68%.

    Parameters
    ----------
                                                                                \n
    p    : polynomial coefficients returned by polyfit2                         \n
    x    : the data for the polynomial evaluation                               \n
    s    : triplet returned by polyfit2                                         \n
                                                                                \n
        (R,df,normr) with R     : the Cholesky factor of the Vandermonde matrix \n
                          df    : the degrees of freedom                        \n
                          normr : the norm of the residuals                     \n
                                                                                \n
    mu   : centering and scaling parameters output by polyfit                   \n
           not accounted here                                                   \n

    Returns
    -------

    y     : the value of the polynomial p evaluated at x                        \n
    delta : estimate of the standard deviation of the error in predicting a     \n
            future observation at X by P(X).                                    \n

    See also
    --------

    polyfit2
    """

    nc = len(p);

    if isscalar(x) and len(args)==0 and nc>0 and isfinite(x) :

       y=lfilter(array([1.]),array([1.,-x]),p);                                 # Make it scream for scalar x.  Polynomial evaluation can be
       y=y[nc-1];                                                               # implemented as a recursive digital filter.

    else:

       if isscalar(x):
          x=array([x]);

       siz_x=len(x);
       if len(args)>=1:
          s=args[0];
       if len(args)>1:
          mu=args[1];
          x=(x-mu[0])/mu[1];

       # Use Horner's method for general case where X is an array.

       y = zeros(siz_x,dtype=double);
       delta = zeros(siz_x,dtype=double);

       if nc>0:
          y=ones(siz_x,dtype=double)*p[0];
       for i in arange(1,nc):
          y=x*y+p[i];

       # Calculate delta

       R=s[0];
       df=s[1];
       normr=s[2];

       # Construct Vandermonde matrix

       V=vander(x,nc);
       E=(lstsq(R.T,V.T))[0].T                                               # E = V/R; (mrdivide function in Matlab)

                                                                             # The NumPy package for Python contains a routine lstsq for computing the least-squares
                                                                             # solution to a system of equations. This routine will likely give you comparable results
                                                                             # to using the MRDIVIDE function in MATLAB, but it is unlikely to be exact.

       e=sqrt(1+sum(E*E,axis=1));                                            # e = sqrt(1+sum(E.*E,2));

       if df==0.:
          printf('[polyval2] Warning : Zero degrees of freedom implies infinite error bounds');
          delta=ones(siz_x,dtype=double)*inf;
       else:
          delta=normr/sqrt(df)*e;

    return y,delta;

def spline(x,y,xnew):
    """
    cubic spline interpolation

    Parameters
    ----------
                                                                                \n
    x    : original grid                                                        \n
    y    : data points defined on the original grid                             \n
    xnew : the new grid

    Returns
    -------

    ynew : the data points interpolated on the new grid
    """

    y2=interpolate.splrep(x,y,k=3,s=0.);                                        # second derivatives calculation (s=0 means no smoothing)
    return interpolate.splev(xnew,y2,der=0);

def interp1(x,y,newx,method):
    """
    interpolates the values of the underlying function y=f(x) at the            \n
    new points newx                                                             \n

    Parameters
    ----------

    x,y    : arrays of values used to approximate some function f               \n
    newx   : the new data points                                                \n
    method : should be 'linear'                                                 \n

    Returns
    -------

    the function calculated in newx
    """

    f=interp1d(x,y,kind=method);
    return f(newx);

def interp2(x,y,mat,newx,newy,method):
    """
    interpolates the values of the underlying function mat=f(x,y) at the        \n
    new points newx and newy                                                    \n

    Parameters
    ----------

    x,y,mat   : arrays of values used to approximate some function f            \n
    newx,newy : the new data points                                             \n
    method : should be 'linear'                                                 \n

    Returns
    -------

    the function calculated in newx
    """

    f = interp2d(x,y,mat,kind=method);
    return f(newx,newy)

def Gamma2sigma(Gamma):
    """
    convert FWHM (Gamma) to standard deviation (sigma)

    Parameters
    ----------

    Gamma : fwhm of the line shape

    Returns
    -------

    sigma : the standard deviation
    """

    return Gamma * sqrt(2) / ( sqrt(2 * log(2)) * 2 )

def Gauss(x,x0,fwhm,asym=0.):
    """
    calculate a gauss line shape with a given fwhm

    Parameters
    ----------

    x    : the grid on which to calculate the Gaussian                          \n
    x0   : central point on the grid                                            \n
    fwhm : the full width at half maximum of the gaussian                       \n

    Returns
    -------

    g : the calculated gaussian
    """

    x1=x/abs(x)

    return exp(-((x-x0)/((1.+x1*asym)*Gamma2sigma(fwhm)))**2 )

def SGauss(x,fwhm,expTerm,asymFactor):
    """
    Calculate a "super gauss" line shape with a given fwhm, exponential term and
    asymmetry factor (cfr Stefen Beirle).  This function is a generalization of
    the Gaussian line shape

    Parameters
    ----------

    x          : the grid on which to calculate the Gaussian                    \n
    fwhm       : the full width at half maximum of the gaussian                 \n
    expTerm    : exponential term                                               \n
    asymFactor : asymmetry factor                                               \n

    Returns
    -------

    g : the calculated line shame
    """

    g=0.;

    if (fwhm<0.):
       print('SGauss : fwhm should be positive')
    elif (expTerm<1.e-14):
       print('SGauss : exponential term is expected to be higher than 0.')
    else:

       a=(fwhm*0.5)/pow(log(2),1./expTerm);
       a2=a*ones(len(x));

       if (abs(asymFactor)>1.e-14):
        for i,item in enumerate(x):
         if item<0.:
          a2[i]=a*(1.-asymFactor);
         else:
          a2[i]=a*(1.+asymFactor);

       w=abs(x);
       g=exp(-pow(w/a2,expTerm));

    return g;

def Lorentz(x,x0,fwhm,n=1):
    """
    calculate a lorentzian line shape with a given fwhm and polynomial order

    Parameters
    ----------

    x    : the grid on which to calculate the Lorentzian                        \n
    x0   : central point on the grid                                            \n
    fwhm : the full width at half maximum of the Lorentzian                     \n

    Returns
    -------

    g : the calculated lorentzian
    """

    sigma=fwhm*0.5;
    return pow(Gamma2sigma(fwhm),n)/(pow(x-x0,n)+pow(Gamma2sigma(fwhm),n));

# NEW IN 2.7 def ErfFunc(x,dx,fwhm):
# NEW IN 2.7
# NEW IN 2.7     """
# NEW IN 2.7     g=ErfFunc(x,dx,fwhm):
# NEW IN 2.7
# NEW IN 2.7     calculate a error function line shape with a given fwhm and pixel width (in nm)
# NEW IN 2.7
# NEW IN 2.7     Parameters
# NEW IN 2.7     ----------
# NEW IN 2.7
# NEW IN 2.7     x    : the grid on which to calculate the error function                    \n
# NEW IN 2.7     dx   : the pixel width                                                      \n
# NEW IN 2.7     fwhm : the full width at half maximum of the gaussian                       \n
# NEW IN 2.7
# NEW IN 2.7     Returns
# NEW IN 2.7     -------
# NEW IN 2.7
# NEW IN 2.7     g : the calculated error function
# NEW IN 2.7     """
# NEW IN 2.7
# NEW IN 2.7     g=zeros(size(x));
# NEW IN 2.7     a=2*sqrt(log(2))/fwhm;
# NEW IN 2.7
# NEW IN 2.7     for i,item in enumerate(x):
# NEW IN 2.7
# NEW IN 2.7         t1=a*(item+dx/2);
# NEW IN 2.7         t2=a*(item-dx/2)
# NEW IN 2.7
# NEW IN 2.7         g[i]=erf(t1)-erf(t2);
# NEW IN 2.7
# NEW IN 2.7     return g;

def scalar_interp(x1,x2,y1,y2,x):
    """
    Linear interpolation between two known points

    Parameters
    ----------

    x1,x2      : coordinates of the two known points                            \n
    y1,y2      : values of the two known points                                 \n
    x          : a coordinate that should be included in x1,x2 interval         \n

    TAKE CARE  : no validity test is performed on x
                 no division by zero test

    Returns
    -------

    y : the interpolated value
    """

    return(y1+(y2-y1)*(x-x1)/(x2-x1));

# -----
# DATES
# -----

def datenum_fromMatlab(matlab_datenum):
    """
    converts the Matlab input date matlab_datenum in a fractional day that corresponds                         \n
    to the number of day since 01/01/0000 where the year 0000 is merely a                                      \n
    reference point and is not intended to be interpreted as a real year.                                      \n
                                                                                                               \n
    Reference : http://sociograph.blogspot.com/2011/04/how-to-avoid-gotcha-when-converting.html                \n

    Parameters
    ----------

    matlab_datenum : result of a datenum in Matlab

    Returns
    -------

    dd : the fractional day
    """

    return datetime.fromordinal(int(matlab_datenum)) + timedelta(days=matlab_datenum%1) - timedelta(days = 366)

def datenum(YYYY,M,D,hh=0,mm=0,ss=0):
    """
    converts the input date (YYYY,M,D,hh,mm,ss) in a fractional day that corresponds
    to the number of day since 01/01/0000 where the year 0000 is merely a
    reference point and is not intended to be interpreted as a real year.

    Parameters
    ----------

    YYYY,M,D,hh,mm,ss : the input date (resp, year, month, day, hour, minute, sec)

    Returns
    -------

    dd : the fractional day
    """

    return datetime(int(YYYY),int(M),int(D)).toordinal()+(hh*3600.+mm*60.+ss)/86400.;

def datevec(dd):
    """
    converts the fractional day (eventually obtained with datenum) in a vector [YYYY,M,D,hh,mm,ss]

    Parameters
    ----------

    dd : the fractional day

    Returns
    -------

    d : the date as a vector (array([YYYY,M,D,hh,mm,ss]))
    """

    dateNum=datetime.fromordinal(int(dd))
    fracHour=(dd-int(dd))*24;
    hvec=dec2dms(fracHour);

    return transpose(array([dateNum.year,dateNum.month,dateNum.day,hvec[0],hvec[1],hvec[2]]));


def fracTime(hh,mm,ss):
    """
    converts the input time (hh,mm,ss) in a fractional hour.

    Parameters
    ----------

    hh,mm,ss : the input time (resp., hour, minute, sec)

    Returns
    -------

    ti : the fractional time
    """

    return hh+mm/60.+ss/3600;

def str2fracTime(timestr):
    """
    converts the input time hh:mm:ss in a fractional hour.

    Parameters
    ----------

    timestr : the input time in hh:mm:ss format

    Returns
    -------

    ti : the fractional time
    """

    [hh,mm,ss] = [float(x) for x in timestr.split(':')]
    return hh+mm/60.+ss/3600;

def getcday(dt):
    """
    converts the input date (array([YYYY,M,D]) in the calendar day [0..365/366]

    Parameters
    ----------

    dt : array([YYYY,M,D]) is the input date

    Returns
    -------

    dd : the calendar day [0..365/366]

    See also
    --------

    datenum
    """

    if (ndim(dt)==1):
       dd=datenum(dt[0],dt[1],dt[2],0,0,0)-datenum(dt[0],1,1,0,0,0)+1;
    else:
       dd=zeros(size(dt,0));
       for i in arange(0,size(dt,0)):
         dd[i]=datenum(dt[i,0],dt[i,1],dt[i,2],0,0,0)-datenum(dt[i,0],1,1,0,0,0)+1;

    return dd;

def getmonth(year,cday):
    """
    return the month in the year from the calendar day

    Parameters
    ----------

    year : given year
    cday : calendar day

    Returns
    -------

    month : the month corresponding to the calendar day in the given year

    See also
    --------

    GetMonths
    """

    # Declarations

    firstDayMonth=array([0,31,59,90,120,151,181,212,243,273,304,334,365,9999],dtype=int);
    month=0;

    # Additional day from February for leap years

    a=0 if (remainder(year,4)!=0) else 1;

    if a==1:
       for i in arange(2,13):
          firstDayMonth[i]=firstDayMonth[i]+1;

    maxday=firstDayMonth[12];

    # Calculate the month in the year

    if (cday<1) or (cday>maxday):
       print("[getmonth] Input calendar day is wrong");
    else:

       while (cday>firstDayMonth[month]):
          month=month+1;

    # Return

    return month;

def GetMonths(years,cdays):
    """
    idem getmonth but applies to vectors

    Parameters
    ----------

    years : array of given years
    cdays : array of given calendar days

    Returns
    -------

    months : the months corresponding to the calendar days in the given years

    See also
    --------

    getmonth
    """

    months=0;

    if len(years)==0 or len(cdays)==0:
       print(sprintf("[GetMonths] Input arrays should contain at least 1 element",len(years),len(cdays)));
    elif len(years)!=len(cdays):
       print(sprintf("[GetMonths] Input arrays should have the same size (%d,%d)",len(years),len(cdays)));
    else :

       months=zeros(len(cdays),dtype=int);

       for j in arange(0,len(cdays)):
           months[j]=getmonth(years[j],cdays[j]);

       return(unique(months));

def getdate(year,cday):
    """
    return the date in array([YYYY,MM,DD]) format from year and calendar day

    Parameters
    ----------

    year : given year
    cday : calendar day

    Returns
    -------

    array([YYYY,MM,DD]) : the date corresponding to the calendar day in the given year

    See also
    --------

    getmonth
    """

    firstDayMonth=array([0,31,59,90,120,151,181,212,243,273,304,334,365,9999],dtype=int);
    month=getmonth(year,cday);
    day=int(cday-firstDayMonth[month-1]);

    # Process leap years

    if (remainder(year,4)==0) and (month>2):
        day=day-1;

    # Return

    return array([year,month,day]);

def MJD2000(dt,ut):
    """
    calculate the modified julian day using 01/01/2000 as reference
    Parameters
    ----------
    dt  : date given as array([YYYY,M,D])                                                      \n
    ut  : universal time given as array([H,M,S]) !!! double brackets for several time          \n
    Returns
    -------
    mjd  : the number of fractional days since 01/01/2000.
    See also
    --------
    datenum
    """
    mjd=datetime(int(dt[0]),int(dt[1]),int(dt[2])).toordinal()-datetime(2000,1,1).toordinal();
    if np.array(ut).size==3:
     return mjd+(ut[0]*3600.+ut[1]*60.+ut[2])/86400;
    else:
     return array([mjd+(x[0]*3600.+x[1]*60.+x[2])/86400. for x in ut]);

# -------
# STRINGS
# -------

def sprintf(format,*args):
    """
    Equivalent to the Matlab sprintf function to write formatted data to string

    Parameters
    ----------

    format : a string that describes the format of the output fields            \n
    *arts : the list of parameters needed to complete the format string
    """

    return(format % args)

# ------------------
# ANGLES CALCULATION
# ------------------

def dec2dms(d):
    """
    converts the decimal angle into degree,minute,seconds

    Parameters
    ----------

    d  : angle to convert in degrees,minutes,seconds

    Returns
    -------

    dd : angle in degree,minute,seconds (array([deg,min,fracSec]))

    See also
    --------

    dms2dec
    """

    if (isscalar(d)):

         deg = fix(d)                                                           # integer part of degrees
         fracMin = fabs((d-deg)*60.)                                            # fractional minutes
         min = fix(fracMin)                                                     # integer part of minutes
         fracSec = int(fabs((fracMin-min)*60.+0.5))                             # fractional seconds

         if (fabs(fracSec-60.)<epsilon):
             fracSec=0.;
             min=min+1;

         if (fabs(min-60.)<epsilon):
             min=0.;
             deg=deg+1;

         ut=transpose(array([deg,min,fracSec]));

    else:

         # Create vector

         ut=zeros([len(d),3],dtype=float);

         # Convert fractional angle in degrees, minutes, seconds

         ut[:,0]=fix(d);
         ut[:,1]=floor((d-floor(d))*60.);
         ut[:,2]=floor(((d-ut[:,0])*60.-ut[:,1])*60.+0.5);

         # Round seconds

         j=[i for (i,item) in enumerate(ut[:,2]) if item==60];
         if (len(j)>0):
            ut[j,2]=0.;
            ut[j,1]=ut[j,1]+1.;

         # Round minutes

         j=[i for (i,item) in enumerate(ut[:,1]) if item==60];
         if (len(j)>0):
            ut[j,1]=0.;
            ut[j,0]=ut[j,0]+1.;

    return ut;

def dms2dec(d):
    """
    converts the angle (given in degree,minute,seconds) in decimal

    Parameters
    ----------

    d  : angle (array([deg,min,fracSec])) to convert in decimal value

    Returns
    -------

    dd : angle in decimal value

    See also
    --------

    dec2dms
    """

    return(dot(d,array([1.,1./60.,1./3600.])));

def sundecl(dt):
    """
    for a given date, calculate the declination of the sun

    Parameters
    ----------

    dt : date given as array([YYYY,M,D])

    Returns
    -------

    d  : sun declination
    """

    dn=getcday(dt);                                                             # get the calculation of the sun
    gma=(2.*pi*(dn-1))/365;
    d=0.006918-0.399912*cos(gma)+0.07257*sin(gma)-0.006758*cos(2.*gma)+0.000907*sin(2.*gma)-0.002697*cos(3.*gma)+0.00148*sin(3.*gma);
    d=d*(180./pi);

    return d;

def loctime(*arg):
    """
    calculate the apparent local time for a given date, longitude and universal time

    Parameters
    ----------

    dt : date given as array of couples (year, calendar day) [YYYY cday] or     \n
         array of dates ([YYYY,M,D])                                            \n
    lg : longitude (+ westwards,- eastwards)                                    \n
    ut : universal time given as array of fractional time or array([H,M,S])     \n

    et : optional, equation of time if known                                    \n

    Returns
    -------

    ld : local day in the same format as dt                                     \n
    lt : apparent local time                                                    \n
    et : equation of time
    """

    dt=arg[0];                                                                  # date given as array([YYYY,M,D])
    lg=arg[1];                                                                  # longitude
    ut=arg[2];                                                                  # universal time as array([YYYY,M,D])

    # if the longitude is given in D,M,S, convert it in decimal value

    if (size(lg)==3):
      lg=dms2dec(lg);

    if (ndim(ut)==2 and size(ut,1)==3):
      hr=dms2dec(ut)-(lg/15.);
    else:
      hr=ut-(lg/15.);

    # only fractional day is provided

    if isscalar(dt):
       dn=floor(dt)*ones((len(ut),));
       ld=c_[zeros((len(dn),)),dn];

    elif (ndim(dt)==1):

       if (dt[0]<=366.):
         dn=floor(dt);
       elif len(dt)==2:
         dn=dt[1]*ones((len(ut),));
       else:
         dn=getcday(dt)*ones((len(ut),));

       ld=c_[zeros((len(dn),)),dn] if dt[0]<=366. else c_[ones((len(dn),))*dt[0],dn];

    # couples year,fractional day are provided

    elif ((ndim(dt)==2) and (size(dt,1)==2)):
       dn=floor(dt[:,1]);
       ld=c_[dt[:,0],dn];

    # [yyyy,mm,dd] are provided

    else:
      dn=getcday(dt);
      ld=c_[dt[:,0],dn];

    # calculation of the equation of time if unknown

    if (len(arg)<4):

      gma=(2.*pi*(dn-1.))/365.;
      et=(0.000075+0.001868*cos(gma)-0.032077*sin(gma)-0.014615*cos(2.*gma)-0.04089*sin(2.*gma))*229.18;

    else:
      et=arg[3];

    # calculation of the apparent local time

    it=hr+(et/60.);

    # correction of dates for high longitudes

    i=[i for (i,item) in enumerate(it) if item>24.];
    if len(i)>0:

       it[i]=it[i]-24.;
       dn[i]=dn[i]+1;

       j=[j for (j,item) in enumerate(dn) if item>365.];
       if len(j)>0:
         for i in j:
            if (mod(ld[i,0],4)!=0) or ((mod(ld[i,0],100)!=0) or (mod(ld[i,0],400)!=0)):
               dn[i]=1;
               ld[i]=ld[i]+1

    i=[i for (i,item) in enumerate(it) if item<0.];
    if len(i)>0:

       it[i]=it[i]+24.;
       dn[i]=dn-1;

       j=[j for (j,item) in enumerate(dn) if item<0];
       if len(j)>0:
         for i in j:
            ld[i]=ld[i]-1;
            dn[i]=365. if (mod(ld[i,0],4)!=0) or ((mod(ld[i,0],100)!=0) or (mod(ld[i,0],400)!=0)) else 366.;

    lt=dec2dms(it) if (ndim(ut)==2) else it;
    ld=c_[ld[:,0],dn];

    # return

    return(lt,ld,et);

def sunangl(*arg):
    """
    calculate the solar zenith and azimuth angles for a given date, longitude,
    latitude and a vector of universal time

    Parameters
    ----------

    dt  : date given as array([YYYY,M,D])                                       \n
    ut  : universal time given as array([[H,M,S]]) !!! double brackets          \n
    lg  : longitude (+ westwards,- eastwards)                                   \n
    lt  : latitude (+ northwards,- southwards)                                  \n
    dlt : optional, declination of the sun                                      \n
    et  : optional, equation of time if known

    Example of usage
    ----------------

    x=arange(735953,735954,30/86400)                                            # in order to have the interval in seconds (in this case, one point every 30 sec)
    hh=(x-735953)*24                                                            # fractional hours
    ut=dec2dms(hh)                                                              # universal time
    (sza,azim)=sunangl([2016,6,21],ut,longit,lat)                               # calculate the solar zenith and azimuth angles

    Returns
    -------

    sza  : the solar zenith angle calculated for the UT                         \n
    azim : the solar azimuth angle calculated for the input UT

    See also
    --------

    sunangl_date
    """

    dt=arg[0];                                                                  # date given as array([YYYY,M,D])
    ut=arg[1];                                                                  # universal time given as array([[H,M,S]])
    lg=arg[2];                                                                  # longitude (+ westwards,- eastwards)
    lt=arg[3];                                                                  # latitude (+ northwards,- southwards)

    # if longitude and latitude are given in D,M,S, convert them in decimal values

    if (size(lg)==3):
        lg=dms2dec(lg);

    if (size(lt)==3):
        lt=dms2dec(lt);

    # calculation of the declination of the sun if unknown

    if (len(arg)<5):
     dlt=sundecl(dt);
    else:
     dlt=arg[4];

    # conversion in radians

    lgr=radians(lg);
    ltr=radians(lt);
    dltr=radians(dlt);

    # Initialization of output

    n=len(ut);

    sza=zeros(n)
    azim=ones(n)*90.;

    # calculate apparent local time

    if (len(arg)<6):
       (loct,ld,et)=loctime(dt,lg,ut);
    else:
       (loct,ld,et)=loctime(dt,lg,ut,et);

    # solar zenith angle

    w=(12.00-dms2dec(loct))*15;
    wr=radians(w);

    sza=degrees(arccos(sin(dltr)*sin(ltr)+cos(dltr)*cos(ltr)*cos(wr)));
    alr=radians(90.-sza);
    
    if isscalar(sza):
       if fabs(lt-90.)<epsilon:
          azim=degrees(arccos(cos(dltr)*cos(wr[0]))/cos(alr));
       else:
          aaa=(sin(alr)*sin(ltr)-sin(dltr))/(cos(alr)*cos(ltr));
          azim=degrees(arccos(aaa));
       if dms2dec(loct)>12.:
          azim=-azim;
       
    else:

     # solar azimuth angle

     alr=radians(90.-sza);

     j=[i for (i,item) in enumerate(sza) if item>=epsilon];

     if len(j)>0:

        if fabs(lt-90.)<epsilon:
           azim[j]=degrees(arccos(cos(dltr)*cos(wr[j]))/cos(alr[j]));
        else:
           aaa=(sin(alr[j])*sin(ltr)-sin(dltr))/(cos(alr[j])*cos(ltr));
           azim[j]=degrees(arccos(aaa));

     j=[i for (i,item) in enumerate(dms2dec(loct)) if item>12.];
     if len(j)>0:
        azim[j]=-azim[j];

    # return

    return (sza,azim)

def isNaN(num):
    return num != num

def sunangl_date(stationName,lg,lt,dt,szaMin,szaMax,plotFlag):
    """
    calculate the solar zenith and azimuth angles at a station for a given date

    Parameters
    ----------

    stationName : the name of the station                                       \n
    lg          : longitude (+ westwards,- eastwards)                           \n
    lt          : latitude (+ northwards,- southwards)                          \n
    dt          : date given as array([YYYY,M,D])                               \n
    szaMin      : exclude data points with SZA < szaMin                         \n
    szaMax      : excude data points with SZA > szaMax

    See also
    --------

    sunangl
    """

    x=arange(735953.,735954.,30./86400.)                                        # in order to have the interval in seconds (in this case, one point every 30 sec)
    hh=(x-735953)*24                                                            # fractional hours
    ut=dec2dms(hh);                                                             # convert fractional hours in deg, min, sec

    # calculate the solar zenith and azimuth angles

    (sza,azim)=sunangl(dt,ut,lg,lt)

    if plotFlag:

       # get time where sza < szaMin

       n1=[i for (i,s) in enumerate(sza) if (s<szaMin)];
       n2=[i for (i,s) in enumerate(sza) if (s<szaMax)];

       # plot SZA

       subplot(2,1,1);
       plot(hh[n2],sza[n2]);
       grid();
       title(sprintf("SZA at %s, %02d/%02d/%d",stationName,dt[2],dt[1],dt[0]),fontsize=14,fontweight='bold');

       if len(n1) :

        axvspan(min(hh[n1]),max(hh[n1]),0,1,hold=False,facecolor='lightgrey');

       # plot azim

       subplot(2,1,2);
       plot(hh[n2],azim[n2]);
       grid();
       title(sprintf("Solar azimuth at %s, %02d/%02d/%d",stationName,dt[2],dt[1],dt[0]),fontsize=14,fontweight='bold');

       if len(n1) :
        axvspan(min(hh[n1]),max(hh[n1]),0,1,hold=False,facecolor='lightgrey');

    return (hh,sza,azim);

def km2deg(x):
    """
    Convert a distance along a great circle of the Earth from km to degrees.
    A radius of 6371 km is assumed.

    Parameters
    ----------

    x : the distance in kms

    Returns
    -------

    a : the distance in degrees
    """

    return(180.*x)/(pi*6371.);

def reckon(lat,lon,arcrange,az):

    """
    For scalar inputs, calculates a position (latout,lonout) at a given range,
    arcrange, and azimuth, az, along a great circle from a starting point defined
    by lat and lon. lat and lon are in degrees. arclen must be expressed as
    degrees of arc on a sphere, and equals the length of a great circle arc
    connecting the point (lat, lon) to the point (latout, lonout). az, also
    in degrees, is measured clockwise from north. reckon calculates multiple
    positions when given four arrays of matching size. When given a combination
    of scalar and array inputs, the scalar inputs are automatically expanded to
    match the size of the arrays.

    Note that this function applies on scalars.  Could be easily extended to
    arrays later.

    (retrieved from the Matlab documentation)

    Returns
    -------

    latout,lonout :
    """

    units = "degrees";
    d=0.;

    if units=="degrees" :
      d=pi/180.;
    elif units=="radians" :
      d=1.;
    else:
     print("[reckon] : unknown unit");

    # Convert to radians

    lat = lat*d;
    lon = lon*d;
    arcrange = arcrange*d;
    az = az*d;

    latout = pi*0.5 - arccos(sin(lat)*cos(arcrange) + cos(lat)*sin(arcrange)*cos(az));

    cos_gamma = (cos(arcrange) - sin(latout)*sin(lat))/(cos(latout)*cos(lat));
    sin_gamma = sin(az)*sin(arcrange)/cos(latout);

    gamma = arctan2(sin_gamma,cos_gamma);

    lonout = lon + gamma;

    # Bring the lonout in the interval [-pi pi]

    lonout = mod(lonout+pi,2.*pi)-pi;

    # Convert to degrees

    lonout = lonout/d;
    latout = latout/d;

    latout=reshape(latout,(len(latout),))
    lonout=reshape(lonout,(len(lonout),))

    # Return

    return (latout,lonout);

# ---------------
# FILES FUNCTIONS
# ---------------

# +++ UTILITIES +++

def loadfile(fileName,c):
    """
    Read ASCII files with comment lines starting with a given character

    Parameters
    ----------

    filename : the name of the ASCII file to read out;                          \n
    c        : the character starting the comment lines

    Returns
    -------

    data     : the matrix of data

    See also
    --------

    loadtxt
    """

    data=loadtxt(fileName, comments=c)
    return data;

# +++ QDOAS +++

def ASCTAB_Read(fileName):
    """
    Read ASCII files generated by QDOAS
    More general, read all ASCII tabbed columns files with columns titles on
    the first line

    Parameters
    ----------

    filename : tabbed-columns ASCII file to read out;

    Returns
    -------

    data     : pandas data frame

    See also
    --------

    read_csv
    """

    data=[];

    if os.path.isfile(fileName):

       if pandasFlag:

          data=DataFrame();

          skip=0;

          f=open(fileName,'r');                                                          # open the file

          for line in f.readlines():
            li=line.lstrip()
            if not li.startswith(";"):
               titleLine=line;
               break;
            skip=skip+1;

          # Retrieve the first line of the file with columns titles

          if ("# Results obtained using Qdoas" in titleLine) or ("# File generated by Qdoas" in titleLine):
             skip=skip+1
          f.close();

          data=read_csv(fileName,header=0,skiprows=skip,sep='\t');

       else:

          # Retrieve the first line of the file with columns titles

          titleLine=f.readline();                                                        # read the first line

          if ("# Results obtained using Qdoas" in titleLine):
             titleLine=f.readline();

          titleColumns=map(str.lower,titleLine.split('\t'));                             # split using tab separator and convert in lower case
          f.close();

          ncol=size(titleColumns);

          # Load the remaining of the file without columns titles (the first line should start with #)

          a=loadtxt(fileName, comments='#', delimiter='\t',usecols=arange(ncol-1))

          # Create dictionary -> do not need ASCTAB_Get anymore but the field should be exact

          if dictFlag:

             data={};

             for i in arange(0,ncol-1):
                data[titleColumns[i]]=a[:,i]

          # Matlab -> for that, need ASCTAB_Get

          else:

            data=[titleColumns,a];
    else:

       print ('File ' + fileName + ' do not exist')

    return data;

# +++ QDOAS netCDF +++

# def QDOAS_Read(fileName,format="netcdf")

# --------------
# DOAS FUNCTIONS
# --------------

def DOAS_DiffCol(sza,scd,scde,sza1,sza2):
    """
    Calculates differential slant columns                                       \n
    A first line is fitted through points arount sza1 and a second one through  \n
    points around sza2.  The fit is weighted by standard deviations using       \n
    polyfit2.                                                                   \n
                                                                                \n
    The differential slant column is calculated as SCD(sza1)-SCD(sza2) and the  \n
    corresponding error is calculated as sqrt(SCDE(sza1)^2+SCDE(sza2)^2).       \n

    Parameters
    ----------

    sza       : solar zenith angles for the considered twilight;                \n
    scd       : slant columns densities for the considered twilight;            \n
    scde      : standard deviations on scd for the considered twilight;         \n
    sza1,sza2 : the reference SZA for calculating DSCDs.

    Returns
    -------

    diffCol   : the differential slant column (sza1-sza2);                      \n
    diffColE  : the error accounting for standard deviations at sza1 and sza2;  \n
    diffP     : the coefficients of the line fitting points around sza1;
    """

    diffCol=double(nan);
    diffColE=double(nan);
    diffP=double(nan);

    isza1=array([i for (i,item) in enumerate(sza) if (item>=sza1-3. and item<=sza1+1.)]);    # isza1=find(sza>=sza1-3 & sza<=sza1+1);
    isza2=array([i for (i,item) in enumerate(sza) if (item>=sza2-2. and item<=sza2+2.)]);    # isza2=find(sza>=sza2-2 & sza<=sza2+2);

    if (len(isza1)>0):

       # get the SZA closest to the higher SZA

       szaDiff=abs(sza[isza1]-sza1);
       ia1=[i for (i,item) in enumerate(szaDiff) if (item==min(szaDiff))];

       scd1=scd[isza1[ia1]];
       scde1=scde[isza1[ia1]];
       scd2=0.;
       scde2=0.;
       diffP=array([0., 0.]);

       # fit a line through the points around sza1

       if len(isza1)>2 and max(sza[isza1])>sza1-0.2:

          diffP,s=polyfit2(sza[isza1],scd[isza1],scde[isza1],1);

          scd1,scde1=polyval2(diffP,sza1,s);

       # fit a line through the points around sza2

       if len(isza2)>2 and min(sza[isza2])<sza2+0.2:

          p,s=polyfit2(sza[isza2],scd[isza2],scde[isza2],1);
          scd2,scde2=polyval2(p,sza2,s);

       # get the DSCD column and error
       diffCol=scd1-scd2;
       diffColE=sqrt(scde1*scde1+scde2*scde2);

    return diffCol,diffColE,diffP

# ----------------
# ARRAYS FUNCTIONS
# ----------------

def get_indexes(a,b):
    """
    Returns the indexes of elements of a that are in b                          \n
    In principle, all elements of b shall be in a but not in the same order     \n

    Parameters
    ----------

    a         : array with elements to filter with elements in b                \n
    b         : array with elements to select in a                              \n

    Returns
    -------

    ia        : the indexes of elements of a that are also in b, order kept     \n
    """

    ia=[];

    for x in b:

       j=[jj for (jj,item) in enumerate(a) if item==x];
       if not len(j):
          print(sprintf('[get_indexes] %g not found in a.',x))
       else:
          ia.append(j[0]);
          if (len(j)>1):
             print(sprintf('[get_indexes] element %g not unique.  Return index of the first one',x))

    return ia;

def get_intersection_indexes(a,b):
    """
    Returns the indexes of common elements in a and b                           \n

    Parameters
    ----------

    a,b       : arrays (expected only one dimension)                            \n
                could be extended to matrices later                             \n

    Returns
    -------

    ia        : the indexes of elements of a that are also in b                 \n
    ib        : the indexes of elements of b that are also in a                 \n
    """

    ia=[];
    ib=[];

    if (len(a)!=len(set(a))):
       print('[get_intersection_indexes] elements of a are not unique')
    elif (len(b)!=len(set(b))):
       print('[get_intersection_indexes] elements of b are not unique')
    else:

       ia=[i for (i,item) in enumerate(a) if item in b];
       ib=ia if not len(ia) else get_indexes(b,a[ia]);

       if (len(ia)!=len(ib)):
          print('[get_intersection_indexes] do not return the same length for indexes arrays')
       elif len(ia)==0:
          print('[get_intersection_indexes] no common elements')

    return (ia,ib)

# -------------
# MISCELLANEOUS
# -------------

def GB_Run_Cmd(s_cmd, tup_args):

    lst_cmd = [s_cmd]
    lst_cmd.extend(tup_args)
    result = sp.call(lst_cmd)
    return result

