'''
Tools to handle netCDF files

Author:  Caroline FAYT (Royal Belgian Institute for Space Aeronomy)
Date:  January 2018
Version:
'''

# ------------------------------------------------------------------------------
# MODULE        netCDF_tools
# ------------------------------------------------------------------------------
# PURPOSE       Tools to handle netCDF files
# ------------------------------------------------------------------------------
# AUTHOR        Caroline FAYT
# COPYRIGHT     Royal Belgian Institute for Space Aeronomy
# CREATION DATE January 2018
#
# MODIFIED
#
# ------------------------------------------------------------------------------

import os;
import netCDF4;

from numpy import *
from datetime import *
from XML_tools import *
import harp

harpFlag=1 if pkgutil.find_loader("harp") else 0
if harpFlag:
   import harp
   
def netCDF_NewVariable(root,varname,vartype,varshape,varunits,vardesc,varfill,value=[],standard_name=''):
    """
    Create a new variable in a netCDF file.  The name, type, shape, units, description and filling value should be provided.

    The content of the variable can already be passed as argument.

    Parameters
    ----------
    root          : the group id the variable should belong to
    varname       : the name of the variable
    vartype       : the type of the variable
    varshape      : the shape of the variable
    varunits      : the units of the variable
    vardesc       : the description of the variable
    varfill       : the filling value
    value         : optional, the content of the variable
    standard_name : optional, standard name if known (to be CF compliant)


    Returns
    -------
    variable : returned by root.createVariable

    See also
    --------
    netCDF4
    """

    variable=root.createVariable(varname,vartype,varshape,zlib=True,fill_value=varfill)
    variable.units=varunits;
    variable.long_name=vardesc;

    if len(standard_name)>0:
       variable.standard_name=standard_name;

    # todo : should check that the dimensions match

    if size(value)>0:
      variable[:]=value;

    return variable;
   
def netCDF_CopyVariable(ncGroupOut,ncGroupIn,varname,newvarname=""):
    """
    Copy a variable from one netCDF file to another one.  A rename of the variable is possible
    
    Parameters
    ----------
    
    ncGroupOut     : the target group id
    ncGroupIn      : the source group id
    varname        : the name of the variable to copy
    newvarname     : optional, the new name in the target group (by default, use the original name in the source group)

    Returns
    -------
    variable : returned by root.createVariable

    See also
    --------
    netCDF4
    """
    
    if len(newvarname)==0:
       newvarname=varname;
      
    varIn=ncGroupIn.variables[varname];
    variable=netCDF_NewVariable(ncGroupOut,newvarname,varIn.datatype,varIn.dimensions, 
                                varIn.getncattr('units') if 'units' in varIn.ncattrs() else '',
                                varIn.getncattr('long_name') if 'long_name' in varIn.ncattrs() else '',
                                varIn.getncattr('_FillValue') if '_FillValue' in varIn.ncattrs() else nan, 
                                value=asarray(varIn),
                                standard_name=varIn.getncattr('standard_name') if 'standard_name' in varIn.ncattrs() else '');
    
    return variable;

def netCDF_GetAttributes(ncFile):
    """
    Returns the list of main attributes of a netCDF file as a dictionary

    Parameters
    ----------
    ncFile      : the name of the netCDF file

    Returns
    -------
    attrList : the list of main attributes of the input netCDF file as a dictionary

    See also
    --------
    netCDF4
    """

    functionName="netCDF_getAttributes"
    attrList={};

    if not os.path.isfile(ncFile):
       print('['+functionName+'] '+ncFile+' not found !!!');
    else:

       try:

           keys=[];

           nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');
           attrs=nc.ncattrs();
           for item in attrs:
               keys.append(nc.getncattr(item));
           nc.close();

           attrList=dict(zip(attrs,keys));

       except OSError :
           print('['+functionName+'] '+'Can not open file '+ncFile+' or attributes not found !!!');

    return attrList;

def netCDF_CopyAttributes(ncFileSource,ncFileTarget,attrList=[]):
    """
    Copy main attributes from one netCDF file to another

    Parameters
    ----------

    ncFileSource   : the name of the source netCDF file
    ncFileTarget   : the name of the target netCDF file
    attrList       : the list of attributes to copy

    Returns
    -------

    0 on success, 1 on error

    See also
    --------

    netCDF4
    """

    functionName="netCDF_copyAttributes"
    rc=1

    if not os.path.isfile(ncFileSource):
       print('['+functionName+'] '+ncFileSource+' not found !!!');
    elif not os.path.isfile(ncFileTarget):
       print('['+functionName+'] '+ncFileTarget+' not found !!!');
    else:

       try:

           ncSource=netCDF4.Dataset(ncFileSource,'r',format='NETCDF4');
           ncTarget=netCDF4.Dataset(ncFileTarget,'r+',format='NETCDF4');

           if len(attrList)==0:
              attrList=ncSource.ncattrs();

           for item in attrList:
              ncTarget.setncattr_string(item,ncSource.getncattr(item));

           ncSource.close();
           ncTarget.close();

           rc=0;

           # sourceAttr=netCDF_GetAttributes(ncFileSource);
           # targetAttr=netCDF_GetAttributes(ncFileTarget);

           # if len(attrList)==0:
           #    attrList=sourceAttr;

           # nc=netCDF4.Dataset(ncFileTarget,'r+',format='NETCDF4');

           # for item in attrList:

           #     if item not in sourceAttr:
           #         print('WARNING : attribute '+item+' not in '+ncFileSource);
           #     else:

           #         if (item in targetAttr) and (sourceAttr[item]!=targetAttr[item]):
           #             print('WARNING : attribute '+item+' modified (' +targetAttr[item]+' -> '+ sourceAttr[item]+')');

           #         nc.setncattr_string(item,sourceAttr[item]);

           # nc.close()

       except OSError:

           print('['+functionName+'] '+'ERROR : can not copy attributes from '+ncFileSource+' to '+ncFileTarget);

       return rc;

def netCDF_Datetime(datestr,timestr):
    """
    Decompose date and time strings in respectively dd/mm/yyyy and hh:mm:ss.ms formats, in seven fields
    yyyy,mm,dd,hh,mm,ss,ms to save in a netCDF file

    Parameters
    ----------
    datestr   : date string in dd/mm/yyyy format
    timestr   : time string in hh:mm:ss.ms format

    Returns
    -------
    a seven values array with the different date and time fields (yyyy,mm,dd,hh,mm,ss,ms)
    """

    # todo : check the validity of the date

    dt=datestr.split('/')
    ut=timestr.split(':');
    ss=ut[2];
    ms='0';

    if '.' in ss:
      ssms=ss.split('.');

      ss=ssms[0];
      ms=ssms[1];

    return asarray([int16(dt[2]),int16(dt[1]),int16(dt[0]),int16(ut[0]),int16(ut[1]),int16(ss),int16(ms)])          # !!! take care of milliseconds for some formats

def netCDF_CheckDimensions(expectedList,currentList,errorFlag=True):
    """
    Check the presence and values of dimensions

    Parameters
    ----------
    expectedList   : dictionary with keynames as the names of dimensions
                                 and values, the expected dimensions values (0 if the dimension size is unknown)
    currentList    : the dimensions retrieved from the netCDF file
    errorFlag      : if an expected dimension is not defined, consider that as an error or a warning

    Returns
    -------
    0 if the expected dimensions are present in the files and have the correct values
    the number of missing or wrong dimensions otherwise
    """

    status="error" if errorFlag==True else "warning"
    rc=0;

    for item in expectedList.keys():

       if errorFlag:
          rc=rc+1;

       if item not in currentList:
          print('[netCDF_CheckDimension] '+status+' : '+item+' dimension not found');
       elif expectedList[item]==0 and currentList[item].size<=0:
          print('[netCDF_CheckDimension] '+status+' : '+item+' dimension should be strictly positive');
       elif expectedList[item]>0 and expectedList[item]!=currentList[item].size:
          print('[netCDF_CheckDimension] '+status+' : '+item+' dimension is invalid ('+str(expectedList[item])+' expected)');
       elif errorFlag:
          rc=rc-1;

    return rc;

def netCDF_CheckGroups(expectedList,currentList,errorFlag=True):
    """
    Check the presence of groups in a netCDF file

    Parameters
    ----------
    expectedList   : the list of expected groups
    currentList    : the list of groups retrieved from the current netCDF file
    errorFlag      : if an expected group is not in the netCDF file, consider that as an error or a warning

    Returns
    -------
    0 if the expected groups are present in the current netCDF file
    the number of missing groups otherwise
    """

    rc=0;

    status="error" if errorFlag==True else "warning"

    for item in expectedList:
       if item not in currentList:
          print('[netCDF_CheckGroup] '+status+' : '+item+' group not found');
          if (errorFlag):
             rc=rc+1;

    return rc;

def netCDF_CheckVariables(expectedList,currentList,currentGroup,errorFlag=True):
    """
    Check the presence of variables in the current group

    Parameters
    ----------
    expectedList   : the list of expected variables
    currentList    : the list of variables retrieved from the current group
    currentGroup   : the name of the current group
    errorFlag      : if an expected variable is not in the current group, consider that as an error or a warning

    Returns
    -------
    0 if the expected variables are present in the current group
    the number of missing variables otherwise
    """

    rc=0;
    status="error" if errorFlag==True else "warning"

    for item in expectedList:
       if item not in currentList:
          print('[netCDF_CheckVariable] '+status+' : '+item+' variable not present in the '+currentGroup+' group');
          if errorFlag:
             rc=rc+1;

    return rc;

def netCDF_CheckDataInRange(varData,varMin,varMax,varFill,varName):
    """
    Check the content of a variable

    Parameters
    ----------
    varData   : the variable to check
    varMin    : the minimum value expected
    varMax    : the maximum value expected
    varFill   : the filling value
    varName   : the name of the variable to check

    Returns
    -------
    0 if the all the values are valid
    """

    rc=1;
    nfill=0;
    nvalid=0;

    ndata=len(varData);

    nvalid=len([i for (i,item) in enumerate(varData) \
                 if (((not isnan(varFill) and item!=varFill) or (isnan(varFill) and not isnan(item))) and \
                      (item>=varMin and item<=varMax))]);
    nfill=len([i for (i,item) in enumerate(varData) \
                 if ((not isnan(varFill) and item==varFill) or \
                     (isnan(varFill) and isnan(item)))]);

    if (nvalid+nfill<ndata):
       print('[netCDF_CheckDataInRange] '+varName+' error : contains values out of range ['+str(varMin)+'..'+str(varMax)+'] '+str(nfill));
    else:

       if (nfill==ndata):
          print('[netCDF_CheckDataInRange] '+varName+' warning : all values are fill value ('+str(varFill)+')');
       rc=0;

    return rc;

def netCDF_CheckDateTime(dt,dtname,d1='',d2=''):
    """
    Check the validity of a date variable

    Parameters
    ----------
    dt   : the table of dates (with 7 columns)
    dtname : the name of the variable in the group

    Returns
    -------
    0 if the all the dates are valid
    """

    rc=1
    ndates=shape(dt)[0];

    if shape(dt)[1] != 7:
       print('[netCDF_CheckDateTime] '+dtname+' : error with dimensions (expected 7 fields)');
    else:
       maxms=0;
       outOfRange=0;
       notIncreasing=0;

       lastDate='00000000T000000Z'

       try:

          for i in arange(0,ndates):
             # Check _FillValue !!!

             dti=dt[i,0:7]
             newDate = datetime(dti[0],dti[1],dti[2],dti[3],dti[4],dti[5])
             if dti[6]>maxms:
                maxms=dti[6];

             dd='%d%02d%02dT%02d%02d%02dZ' % (dti[0],dti[1],dti[2],dti[3],dti[4],dti[5]);
             if dd>lastDate:
                lastDate=dd;
             elif dd<lastDate:
                notIncreasing=notIncreasing+1;

          rc=0 # Consider the following issues as warnings
               # Could be changed into error later

          if (maxms>0 and maxms<100):
             print('[netCDF_CheckDateTime warning] '+dtname+' : milliseconds could be hundreth of seconds errors');
          elif (notIncreasing):
             print('[netCDF_CheckDateTime warning] '+dtname+' : contains not increasing date and time');
          else:
             rc=0;

       except ValueError as e:
          print('[netCDF_CheckDateTime error] '+dtname+' : contains errors ('+str(e)+')');

    return rc;

def netCDF_SaveAttributes_toXML(nc,f,indent):
    """
    Save attributes of the current netCDF object to XML

    Parameters
    ----------
    nc       : the netCDF group
    f        : pointer to the output file (XML format)
    indent   : current identation
    """

    functionName="netCDF_SaveAttributes_toXML"

    try:

       # Save attributes

       f.write(indent+'<ATTRIBUTES>\n');
       for attr in nc.ncattrs():
          f.write(indent+'   <'+attr+' mandatory="True" empty_allowed="False" expected_value="'+nc.getncattr(attr)+'" />\n');
       f.write(indent+'</ATTRIBUTES>\n');

    except BaseException as error:
       print('['+functionName+'] returned fatal error : {}'.format(error))

    return;

def netCDF_SaveDimensions_toXML(nc,f,indent):
    """
    Save dimensions of the current netCDF object to XML

    Parameters
    ----------
    nc       : the netCDF group
    f        : pointer to the output file (XML format)
    indent   : current identation
    """

    functionName="netCDF_SaveDimensions_toXML"

    try:

       # Save dimensions

       f.write(indent+'<DIMENSIONS>\n');
       for dim_name in nc.dimensions:
             f.write(indent+'   <'+dim_name+' mandatory="True" size="0" />\n');   # use size 0 for variable dimension
       f.write(indent+'</DIMENSIONS>\n');

    except BaseException as error:
       print('['+functionName+'] returned fatal error : {}'.format(error))

    return;

def netCDF_SaveVariables_toXML(nc,f,indent):
    """
    Save dimensions of the current netCDF object to XML

    Parameters
    ----------
    nc       : the netCDF group
    f        : pointer to the output file (XML format)
    indent   : current identation
    """

    functionName="netCDF_SaveVariables_toXML"

    try:

       # Save dimensions

       f.write(indent+'<VARIABLES>\n');
       for var_name in nc.variables:
          f.write(indent+'   <'+var_name+' ');
          var_data=nc.variables[var_name];
          for attr in var_data.ncattrs():
             f.write(attr+'="'+str(var_data.getncattr(attr))+'" ');
          f.write('ndim="'+str(var_data.ndim)+'" dimensions="'+str(var_data.dimensions)+'" mandatory="True" min="0." max="0." />\n');
       f.write(indent+'</VARIABLES>\n');

    except BaseException as error:
       print('['+functionName+'] returned fatal error : {}'.format(error))

    return;

def netCDF_SaveGroup_toXML(ncGroup,f,name,indent):
    """
    Save information on the group to XML

    Add mandatory flag (initialized to True but could be changed later)

    Parameters
    ----------
    ncGroup  : the netCDF group
    f        : pointer to the output file (XML format)
    name     : the name of the nc object to output in the XML file
    indent   : current identation
    """

    functionName="netCDF_SaveGroup_toXML"

    nattrs=len(ncGroup.ncattrs());
    ndims=len(ncGroup.dimensions);
    ngroups=len(ncGroup.groups);
    nvars=len(ncGroup.variables);

    f.write(indent+'<'+name+' nattrs="'+str(nattrs)+'" ndims="'+str(ndims)+'" ngroups="'+str(ngroups)+'" nvars="'+str(nvars)+'"');
    if len(indent)>0:
       f.write(' mandatory="True" >\n');
    else:
       f.write(' >\n');


    new_indent=indent+'   '

    if nattrs>0:
       netCDF_SaveAttributes_toXML(ncGroup,f,new_indent);
    if ndims>0:
       netCDF_SaveDimensions_toXML(ncGroup,f,new_indent);
    if ngroups>0:
       f.write(new_indent+'<GROUPS>\n');
       for item in ncGroup.groups:
          netCDF_SaveGroup_toXML(ncGroup.groups[item],f,item,new_indent+'   ')
       f.write(new_indent+'</GROUPS>\n');
    if nvars>0:
       netCDF_SaveVariables_toXML(ncGroup,f,new_indent)

    f.write(indent+'</'+name+'>\n');

def netCDF_SaveToXML(ncFile,xmlFile):
    """
    Save the structure of a netCDF file in a XML file for further comparison

    Parameters
    ----------
    ncFile   : the name of the input file (netCDF format)
    xmlFile  : the name of the output file (XML format)
    """

    functionName="netCDF_SaveToXML"

    if not os.path.isfile(ncFile):
       print ('['+functionName+'] '+'File ' + ncFile + ' do not exist')
    elif not os.path.isdir(os.path.dirname(xmlFile)):
       print ('['+functionName+'] '+'Can not create '+xmlFile)
    else:

       try:

         nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');

         f=open(xmlFile,'w');
         netCDF_SaveGroup_toXML(nc,f,'netCDF_file_description','');
         f.close();

         nc.close();

       except OSError:
           print('['+functionName+'] '+'ERROR : can not read netCDF file '+ncFile+' or can not open '+xmlFile+' in writing mode');

def netCDF_CompareAttributes_toXML(rootName,ncGroup,xmlGroup,indent):
    """
    Check if attributes in the XML group are present in the netCDF group

    Parameters
    ----------
    rootName : the current name
    ncGroup  : the netCDF group
    xmlGroup : description of the group in the XML file
    indent   : current indentation

    Returns
    -------
    0 on success, any other value on error
    """

    functionName="netCDF_CompareAttributes_toXML"

    new_indent=indent+'   '
    rc=0

    for attr in xmlGroup.ATTRIBUTES:

       newName=rootName+'.'+attr;

       if attr in ncGroup.ncattrs():

          if len(xmlGroup.ATTRIBUTES[attr].expected_value)>0 :

             expected_value=xmlGroup.ATTRIBUTES[attr].expected_value;
             nc_value=ncGroup.getncattr(attr);

             if xmlGroup.ATTRIBUTES[attr].empty_allowed=="False" and len(nc_value)==0:
                print(new_indent+'Error : attribute '+newName+' is empty');
                rc=rc+1;
             elif ('|' not in expected_value) and (nc_value!=expected_value):
                print(new_indent+'Warning : attribute '+newName+' expected value should be '+expected_value+' (currently : '+nc_value+') ');
             elif ('|' in expected_value) and (ncGroup.getncattr(attr) not in expected_value.split('|')) :
                print(new_indent+'Error : attribute '+newName+' expected value not in '+str(expected_value.split('|'))+' list'+' (currently : '+nc_value+') ');
                rc=rc+1;

       elif xmlGroup.ATTRIBUTES[attr].mandatory=="False":
          print(new_indent+'Warning : attribute '+newName+' not present but not mandatory');
       else:
          print(new_indent+'Error : attribute '+newName+' not present');
          rc=rc+1;

    return rc;

def netCDF_CompareDimensions_toXML(rootName,ncGroup,xmlGroup,indent):
    """
    Check if dimensions in the XML group are present in the netCDF group

    Parameters
    ----------
    rootName : the current name
    ncGroup  : the netCDF group
    xmlGroup : description of the group in the XML file
    indent   : current indentation

    Returns
    -------
    0 on success, any other value on error
    """

    functionName="netCDF_CompareDimensions_toXML"

    new_indent=indent+'   '
    rc=0;

    for dim in xmlGroup.DIMENSIONS:

          newName=rootName+'.dimensions[\''+dim+'\']';

          if dim in ncGroup.dimensions:

             if int(xmlGroup.DIMENSIONS[dim].size)>0 and  ncGroup.dimensions[dim].size!=int(xmlGroup.DIMENSIONS[dim].size):
                print(new_indent+'Error : '+xmlGroup.DIMENSIONS[dim].size+' is expected for dimension '+newName+'(currently : '+int(xmlGroup.DIMENSIONS[dim].size)+')');
                rc=rc+1;

          elif xmlGroup.DIMENSIONS[dim].mandatory=="False":
             print(new_indent+'Warning : dimension '+newName+' not present but not mandatory');
          else:
             print(new_indent+'Error : dimension '+newName+' not present');
             rc=rc+1;

    return rc;

def netCDF_CompareVariables_toXML(rootName,ncGroup,xmlGroup,indent):
    """
    Check if variables in the XML group are present in the netCDF group

    Parameters
    ----------
    rootName : the current name
    ncGroup  : the netCDF group
    xmlGroup : description of the group in the XML file
    indent   : current indentation

    Returns
    -------
    0 on success, any other value on error
    """

    functionName="netCDF_CompareVariables_toXML"

    new_indent=indent+'   '
    rc=0;

    for var in xmlGroup.VARIABLES:

          newName=rootName+'.variables[\''+var+'\']';

          if var in ncGroup.variables :

             nc_var=ncGroup.variables[var];
             xml_var=xmlGroup.VARIABLES[var];

             for key in xml_var.keys():

                if key!="mandatory" and key!="ndim" and key!="dimensions" and key!="min" and key!="max" :

                   if key not in nc_var.ncattrs() :
                      print(new_indent+'Warning : variable '+newName+' do not contain attribute '+key+'='+xml_var[key]);
                   else:

                      nc_attr=str(nc_var.getncattr(key));
                      xml_attr=xml_var[key];

                      if nc_attr.lower()!=xml_attr.lower():
                         print(new_indent+'Warning : variable '+newName+'.'+key+' expected value should be '+xml_var[key]+' (currently : '+str(nc_var.getncattr(key))+') ');

                elif key=="ndim" and nc_var.ndim==0:
                   print(new_indent+'Warning : variable '+newName+'.ndim'+' is zero.  Value "'+xml_var[key]+'" expected (dimensions : '+xml_var["dimensions"]+')');
                elif key=="ndim" and xml_var[key]!=str(nc_var.ndim):
                   print(new_indent+'Error : variable '+newName+'.ndim'+' expected value should be '+xml_var[key]+' (currently : '+str(nc_var.ndim)+') ');
                   rc=rc+1;
                elif key=="dimensions" and nc_var.ndim!=0 and xml_var[key]!=str(nc_var.dimensions):
                   print(new_indent+'Error : variable '+newName+'.dimensions'+' expected value should be '+xml_var[key]+' (currently : '+str(nc_var.dimensions)+') ');
                   rc=rc+1;

             var_min=float(xml_var["min"] if "min" in xml_var.keys() else "0.");
             var_max=float(xml_var["max"] if "max" in xml_var.keys() else "0.");
             var_fill=float(xml_var["_FillValue"] if "_FillValue" in xml_var.keys() else "nan");

             if (var_min>var_max):
                var_min,var_max=var_max,var_min;

             if var_max-var_min>1.e-14:

                if (nc_var.ndim>0) and (size(nc_var)>0):
                 
                   # RuntimeWarning: invalid value encountered in less (or great) -> presence of nan values

                   print(var,var_fill,var_min,var_max);
                   i=[i for (i,item) in enumerate(reshape(asarray(nc_var),(size(nc_var),))) if (item!=var_fill) and ((item<var_min) or (item>var_max))];

                   if len(i)>0:
                      print(new_indent+'Error : variable '+newName+' contains value out of ['+str(var_min)+'-'+str(var_max)+']');
                      rc=rc+1;

                elif nc_var!=var_fill and (float(asarray(nc_var))<var_min or float(asarray(nc_var))>var_max):
                      print(new_indent+'Error : variable '+newName+' contains value out of ['+str(var_min)+'-'+str(var_max)+']');
                      rc=rc+1;

          elif xmlGroup.VARIABLES[var].mandatory=="False":
             print(new_indent+'Warning : variable '+newName+' not present but not mandatory');
          else:
             print(new_indent+'Error : variable '+newName+' not present');
             rc=rc+1;

    return rc;

def netCDF_CompareGroup_toXML(rootName,ncGroup,xmlGroup,indent,rc):
    """
    Check if information in the XML group are present in the netCDF group

    Parameters
    ----------
    rootName : the current name
    ncGroup  : the netCDF group
    xmlGroup : description of the group in the XML file
    indent   : current indentation
    rc       : current return code (addition of errors)

    Returns
    -------
    the modified return code
    """

    functionName="netCDF_CompareGroup_toXML"

    new_indent=indent+'   '

    if int(xmlGroup.nattrs)>0:
       rc=rc+netCDF_CompareAttributes_toXML(rootName,ncGroup,xmlGroup,new_indent);
    if int(xmlGroup.ndims)>0:
       rc=rc+netCDF_CompareDimensions_toXML(rootName,ncGroup,xmlGroup,new_indent);
    if int(xmlGroup.ngroups)>0:

       for group in xmlGroup.GROUPS:

          newName=rootName+'.groups[\''+group+'\']'
          print(new_indent+newName);

          if group in ncGroup.groups:
             rc=rc+netCDF_CompareGroup_toXML(newName,ncGroup.groups[group],xmlGroup.GROUPS[group],new_indent,0);
          elif xmlGroup.GROUPS[group].mandatory=="False":
             print(new_indent+'Warning : group '+newName+' not present but not mandatory');
          else:
             print(new_indent+'Error : group '+newName+' not present');
             rc=rc+1

    if int(xmlGroup.nvars)>0:
       rc=rc+netCDF_CompareVariables_toXML(rootName,ncGroup,xmlGroup,new_indent);

    return rc;

def netCDF_CompareToXML(ncFile,xmlFile):
    """
    Compare the structure of a netCDF file to a template one saved in a XML file (created by netCDF_SaveToXML)

    Parameters
    ----------
    ncFile   : the name of the input file (netCDF format)
    xmlFile  : the name of the XML file (XML format)

    Returns
    -------
    the modified return code
    """

    functionName="netCDF_CompareToXML"
    rc=0;

    if not os.path.isfile(ncFile):
       print ('['+functionName+'] '+'File ' + ncFile + ' do not exist')
    elif not os.path.isfile(xmlFile):
       print ('['+functionName+'] '+'File ' + xmlFile + ' do not exist')
    else:

       try:

         nc=netCDF4.Dataset(ncFile,'r',format='NETCDF4');
         l1_structure=XML_LoadStructure(xmlFile);
         print(os.path.basename(ncFile));
         rc=netCDF_CompareGroup_toXML('root',nc,l1_structure,'',0);

       except OSError:
           print('['+functionName+'] '+'ERROR : can not read netCDF file '+ncFile+ ' or XML template '+xmlFile);
           rc=-1;

    return rc;

def netCDF_ReadQDOAS(qdoasFile):
    """
    Read a QDOAS output file in netCDF format

    Parameters
    ----------
    qdoasFile   : the name of a file generated by QDOAS

    Returns
    -------
    a structure similar to the one returned by ASCTAB_Read
    """

    functionName="netCDF_ReadQDOAS"
    data={};

    if not os.path.isfile(qdoasFile):
       print ('['+functionName+'] '+'File ' + qdoasFile + ' do not exist')
    else:

       try:

          # test if it is possible to open file

          nc=netCDF4.Dataset(qdoasFile,'r',format='NETCDF4');

          if len(nc.groups)==0:
             print('['+functionName+'] '+qdoasFile+' is empty !!!');
          elif len(nc.groups)!=1 :
             print('['+functionName+'] '+qdoasFile+' is not a QDOAS file !!!');
          else:

             qdoas_name=list(nc.groups.keys())[0];
             qdoas_group=nc.groups[qdoas_name];

             for qdoas_var in qdoas_group.variables:
                 data[qdoas_var]=asarray(qdoas_group.variables[qdoas_var]);

             # Browse analysis windows

             for analysis_window in qdoas_group.groups:
                 for qdoas_var in qdoas_group[analysis_window].variables:
                     data[analysis_window+'.'+qdoas_var]=asarray(qdoas_group[analysis_window].variables[qdoas_var])

          nc.close();

       except OSError:

           print('['+functionName+'] '+'ERROR : can not read netCDF file '+qdoasFile);

    return data

if harpFlag:

   class harp_qdoas(dict) :

      # Create the structure

      def __init__(self):

         self.__dict__ = self
         self.harp_product=harp.Product(source_product="QDOAS slant columns densities");

         pass;

      # Add variable

      def AddVariable(self,name, \
                      data=[],\
                      unit='', \
                      valid_min=0.0, \
                      valid_max=0.0,\
                      fill_value=0.0, \
                      description='', \
                      enum=[]):

          rc=1;

          if len(name)==0 :
             print('Name of variable is empty');
          elif name.find(' ')>=0:
             print('Name of variable "'+name+' is invalid');
          elif len(data)==0:
             print('No data provided for variable '+name);
          else:

             rc=0;

             if (abs(fill_value)>1.e-6):
              data[data==fill_value]=valid_max*10;

             if (ndim(data)==1):
                setattr(self.harp_product,name,harp.Variable(data,["time"]));
             else:
                setattr(self.harp_product,name,harp.Variable(data,["time",None]));

             setattr(self.harp_product[name],"unit",unit);                                 # physical unit the values associated with the variable are expressed in;
             setattr(self.harp_product[name],"description",description);                   # human-readable description of the variable;

             if abs(valid_max-valid_min)>1.e-6:
                setattr(self.harp_product[name],"valid_min",asscalar(asarray(valid_min,dtype=data.dtype)));    # minimum valid value; any value below this threshold is considered to be invalid;
                setattr(self.harp_product[name],"valid_max",asscalar(asarray(valid_max,dtype=data.dtype)));    # maximum valid value; any value above this threshold is considered to be invalid;
             if len(enum)>0:
                setattr(self.harp_product[name],"enum",enum);                              # list of strings with the names of each enumeration value.

          return rc;

   HARP_species_dict={
    "BrO":"bromine oxide",
    "C2H2O2":"glyoxal",
    "HCHO":"formaldehyde",
    "H2O":"water",
    "IO":"hypoiodite",
    "NO2":"nitrogen dioxide",
    "OClO":"chlorine dioxide",
    "O3":"ozone",
    "O4":"tetraoxygen, oxozone",
    "SO2":"sulfur dioxide",
    };

   QDOAS_satellite_sensors=[ "GOME", "GOME-2","SCIAMACHY","OMI","TROPOMI","S5P" ];
   QDOAS_groundbased_sensors=[ "PANDORA", "UVVIS-DOAS","UVVIS-ZENITH","UVVIS-OFFAXIS","UVVIS-DIRECTSUN" ];

   QDOAS_fields_dict={

    "Spec No" :
       { "harp_name" : "index",
         "unit" : "",
         "description" : "zero-based index of the sample within the source product",
         "valid_min" : 0.,
         "valid_max" : 0.,
         "type"      : int32,
         "correction" : -1 },

    "SZA" :
       { "harp_name" : "solar_zenith_angle",
         "unit" : "degree",
         "description" : "solar zenith angle",
         "valid_min" : 0.,
         "valid_max" : 180.,
         "type"      : double,
         "correction" : 0. },

    "Solar Azimuth Angle" :
       { "harp_name" : "solar_azimuth_angle",
         "unit" : "degree",
         "description" : "solar azimuth angle (0-360, 0 degree North)",
         "valid_min" : 0.,
         "valid_max" : 360.,
         "type"      : double,
         "correction" : 0. },

    "Elev. viewing angle" :
       { "harp_name" : "viewing_elevation_angle",
         "unit" : "degree",
         "description" : "viewing elevation angle",
         "valid_min" : 0.,
         "valid_max" : 180.,
         "type"      : double,
         "correction" : 0. },

    "Tint" :
       { "harp_name" : "integration_time",
         "unit" : "s",
         "description" : "integration time",
         "valid_min" : 0.,
         "valid_max" : 0.,
         "type"      : double,
         "correction" : 0. },

    "Latitude" :
       { "harp_name" : "latitude",
         "unit" : "degree_north",
         "description" : "latitude",
         "valid_min" : -90.,
         "valid_max" : 90.,
         "type"      : double,
         "correction" : 0. },

    "Longitude" :
       { "harp_name" : "longitude",
         "unit" : "degree_east",
         "description" : "longitude",
         "valid_min" : -180.,
         "valid_max" : 360.,
         "type"      : double,
         "correction" : 0. },

    "Pixel corner latitudes" :
       { "harp_name" : "latitude_bounds",
         "unit" : "degree_north",
         "description" : "latitudes of the ground pixel corners",
         "valid_min" : -90.,
         "valid_max" : 90.,
         "type"      : double,
         "correction" : 0. },

    "Pixel corner longitudes" :
       { "harp_name" : "longitude_bounds",
         "unit" : "degree_east",
         "description" : "longitudes of the ground pixel corners",
         "valid_min" : -180.,
         "valid_max" : 360.,
         "type"      : double,
         "correction" : 0. },

    "Satellite height" :
       { "harp_name" : "sensor_altitude",
         "unit" : "m",
         "description" : "altitude of the satellite with respect to the geodetic sub-satellite point",
         "valid_min" : 0.,
         "valid_max" : 0.,
         "type"      : double,
         "correction" : 0. },

    "Satellite latitude" :
       { "harp_name" : "sensor_latitude",
         "unit" : "degree_north",
         "description" : "latitude of the geodetic sub-satellite point",
         "valid_min" : -90.,
         "valid_max" : 90.,
         "type"      : double,
         "correction" : 0. },

    "Satellite longitude" :
       { "harp_name" : "sensor_longitude",
         "unit" : "degree_east",
         "description" : "longitude of the geodetic sub-satellite point",
         "valid_min" : -180.,
         "valid_max" : 360.,
         "type"      : double,
         "correction" : 0. },

    "Azim. viewing angle" :
       { "harp_name" : "viewing_azimuth_angle",
         "unit" : "degree",
         "description" : "viewing azimuth angle",
         "valid_min" : 0.,
         "valid_max" : 360.,
         "type"      : double,
         "correction" : 0. },

    "Viewing zenith angle" :
       { "harp_name" : "viewing_zenith_angle",
         "unit" : "degree",
         "description" : "line of sight zenith angle at top of atmosphere",
         "valid_min" : -90.,
         "valid_max" : 90.,
         "type"      : double,
         "correction" : 0. },

    "Viewing zenith angle at satellite" :
       { "harp_name" : "sensor_viewing_zenith_angle",
         "unit" : "degree",
         "description" : "viewing zenith angle at satellite",
         "valid_min" : -90.,
         "valid_max" : 90.,
         "type"      : double,
         "correction" : 0. },

    "LoS ZA" :
       { "harp_name" : "viewing_zenith_angle",
         "unit" : "degree",
         "description" : "line of sight zenith angle at the ground pixel location",
         "valid_min" : -90.,
         "valid_max" : 90.,
         "type"      : double,
         "correction" : 0. },

    "LoS Azimuth" :
       { "harp_name" : "viewing_azimuth_angle",
         "unit" : "degree",
         "description" : "azimuth angle of the satellite at the ground pixel location",
         "valid_min" : 0.,
         "valid_max" : 360.,
         "type"      : double,
         "correction" : 0. },

    "Solar zenith angle at satellite" :
       { "harp_name" : "sensor_zenith_angle",
         "unit" : "degree",
         "description" : "zenith angle of the satellite at the ground pixel location",
         "valid_min" : -90.,
         "valid_max" : 90.,
         "type"      : double,
         "correction" : 0. },

    "Solar azimuth angle at satellite" :
       { "harp_name" : "sensor_azimuth_angle",
         "unit" : "degree",
         "description" : "azimuth angle of the satellite at the ground pixel location",
         "valid_min" : 0.,
         "valid_max" : 360.,
         "type"      : double,
         "correction" : 0. },

    "Pixel type" :
       { "harp_name" : "scan_subindex",
         "unit" : "",
         "description" : "relative index of this measurement within a scan (forward + backward)",
         "valid_min" : 0,
         "valid_max" : 32767,
         "type"      : int8,
         "correction" : 0 },

    "Cloud fraction":
       { "harp_name" : "cloud_fraction",
         "unit" : "",
         "description" : "cloud fraction",
         "valid_min" : 0.,
         "valid_max" : 1.,
         "type"      : double,
         "correction" : 0. },

    "Cloud Top Pressure" :
       { "harp_name" : "cloud_top_pressure",
         "unit" : "hPa",
         "description" : "cloud top pressure",
         "valid_min" : 0.,
         "valid_max" : 0.,
         "type"      : double,
         "correction" : 0. },

    "Orbit number" :
       { "harp_name" : "orbit_index",
         "unit" : "",
         "description" : "absolute orbit index",
         "valid_min" : 0,
         "valid_max" : 0,
         "type"      : int32,
         "correction" : 0. }
      };

   def netCDF_fromQDOAS_GetDatetime(qdoas_var,hdim,ms_factor=1.):

       qdoas_datetime=array(zeros(hdim));

       # case 1 : use "Date & time (YYYYMMDDhhmmss)"

       if 'Date & time (YYYYMMDDhhmmss)' in qdoas_var :

          dt=asarray(qdoas_var['Date & time (YYYYMMDDhhmmss)'])
          dt=reshape(dt,(hdim,size(dt,-1)))

          fillValue=qdoas_var['Date & time (YYYYMMDDhhmmss)']._FillValue;
          dtok=where(dt[:,0]!=fillValue);

          qdoas_datetime[where(dt[:,0]==fillValue)]=fillValue; # dtnok
          qdoas_datetime[dtok]=(array([MJD2000(array([dt[x,0],dt[x,1],dt[x,2]]),array([dt[x,3],dt[x,4],dt[x,5]])) for x in dtok[0]]))*86400.+dt[dtok,6]*0.001*ms_factor;

       # Case 2 : use "Fractional day" (Year should be specified)

       elif 'Year' in qdoas_var and 'Fractional day' in qdoas_var:
          year=asarray(qdoas_var['Year']).reshape(hdim);
          dd=asarray(qdoas_var['Fractional day']).reshape(hdim);
          qdoas_datetime=(array([MJD2000(array([y,1,1]),array([0,0,0])) for y in year])+dd-1)*86400.;

       # Case 3 : use "Fractional time"

       elif 'Year' in qdoas_var and \
           ('Day number' in qdoas_var or 'Fractional day' in qdoas_var) and \
            'Fractional time' in qdoas_var :

          year=asarray(qdoas_var['Year']).reshape(hdim);
          dd=asarray(qdoas_var['Day number' in qdoas_var]).reshape(hdim) if 'Day number' in qdoas_var else \
             floor(asarray(qdoas_var['Fractional day']).reshape(hdim));
          qdoas_datetime=(array([MJD2000(array([y,1,1]),array([0,0,0])) for y in year])+dd-1)*86400.+ asarray(qdoas_var['Fractional time'])*3600.;

       else:

          print('Impossible to build datetime variable');

       return qdoas_datetime;


   def netCDF_fromQDOAS_toHarp(qdoasFile,harpFile,species_dict={}, \
                               sensor_name="",                     \
                               sensor_latitude=None,               \
                               sensor_longitude=None,              \
                               sensor_altitude=None,               \
                               azimuth_correction=0.):
       """
       Read a QDOAS output file (in netCDF format) and convert it in a new netCDF file
       following the HARP convention

       Parameters
       ----------

       qdoasFile   : the name of a file generated by QDOAS
       harpFile    : the name of the output file (HARP convention)

       azimuth_correction : different conventions exist for azimuth
                            this parameter allows to convert the azimuth angles from the user convention to 0-360 range, North-East convetion

       Other options :

          for Ground-based :

               sensor_name       : the name of the sensor
               sensor_latitude   : the latitude of the station
               sensor_longitude  : the longitude of the station
               sensor_altitude   : the altitude of the station
       """

       functionName="netCDF_fromQDOAS_toHarp"

       hp=harp_qdoas();

       if not os.path.isfile(qdoasFile):
          print ('['+functionName+'] '+'File ' + qdoasFile + ' do not exist')
       elif not os.path.dirname(harpFile):
          print ('['+functionName+'] '+'directory ' + os.path.dirname(harpFile) + ' do not exist')
       else:

          try :

             # test if it is possible to open file
             nc=netCDF4.Dataset(qdoasFile,'r',format='NETCDF4');

             print('QDOAS file : '+qdoasFile);

             if len(nc.groups)==0:
                print('['+functionName+'] '+qdoasFile+' is empty !!!');
             elif len(nc.groups)!=1 :
                print('['+functionName+'] '+qdoasFile+' is not a QDOAS file !!!');
             else:

                qdoas_name=list(nc.groups.keys())[0];
                qdoas_group=nc.groups[qdoas_name];
                qdoas_var=qdoas_group.variables;

                satelliteFlag=1 if len(sensor_name)>0 and sensor_name in QDOAS_satellite_sensors else 0;

                # Help : print the content of the QDOAS file

                if len(species_dict)==0:

                   for analysis_window in qdoas_group.groups:
                      for column in qdoas_group.groups[analysis_window].variables:
                         if ("SlCol" in column):
                            species=column[6:-1];
                            if (species[0]!='x') and (species!="resol"):
                               print('      ("'+analysis_window+'","'+column[6:-1]+'")');

                elif len(qdoas_var)==0:
                   printf('['+functionName+'] '+qdoasFile+' do not contain variables !!!');
                else:

                   if (azimuth_correction>0.) :

                      QDOAS_fields_dict["Solar Azimuth Angle"]["correction"]=azimuth_correction;
                      QDOAS_fields_dict["Azim. viewing angle"]["correction"]=azimuth_correction;
                      QDOAS_fields_dict["LoS Azimuth"]["correction"]=azimuth_correction;
                      QDOAS_fields_dict["Solar azimuth angle at satellite"]=azimuth_correction;

                   qdoas_keys=qdoas_var.keys()
                   xdim=0;

                   for k in qdoas_keys:
                      xdim=len(asarray(qdoas_var[k]));
                      break;

                   xdim=xdim if "n_alongtrack" not in qdoas_group.dimensions else qdoas_group.dimensions["n_alongtrack"].size;
                   ydim=1 if "n_crosstrack" not in qdoas_group.dimensions else qdoas_group.dimensions["n_crosstrack"].size;
                   hdim=xdim*ydim;
                   nhdim=1 if ydim==1 else 2;

                   if (ydim>1):
                      scan_subindex=asarray(arange(ydim)*ones((xdim,ydim)),dtype=int16).reshape(hdim);
                      hp.AddVariable("scan_subindex",
                                     data=asarray(scan_subindex).reshape(hdim),
                                     description="zero-based index of the pixel within the scanline");

                   # Build date time

                   ms_factor=1. if satelliteFlag==0 else 1.e-3;

                   dd=netCDF_fromQDOAS_GetDatetime(qdoas_var,hdim,ms_factor=ms_factor);
                   if len(dd)>0:
                      hp.AddVariable("datetime",
                                     data=asarray(dd),
                                     unit="s since 2000-01-01",
                                     description="time of measurement, number of seconds since 2000-01-01");

                   # Output geodata

                   for varname in qdoas_var:
                      if varname in QDOAS_fields_dict:
                         qdoas_field=QDOAS_fields_dict[varname];

                         qdoas_data=asarray(qdoas_var[varname]);

                         if ndim(qdoas_data)>nhdim:

                            qdoas_data=reshape(qdoas_data,(hdim,size(qdoas_data,-1)))

                            if sensor_name=="SCIAMACHY" and \
                              (varname=="SZA") or (varname=="Solar Azimuth Angle") or \
                              (varname=="LoS ZA") or (varname=="LoS Azimuth"):
                               qdoas_data=qdoas_data[:,1];


                         else:
                            qdoas_data=asarray(qdoas_var[varname]).reshape(hdim);

                         # Reorganize longitudes/latitudes bounds according to the sensor

                         if (varname=="Pixel corner latitudes" or varname=="Pixel corner longitudes") and satelliteFlag:
                            if sensor_name=="GOME":
                               qdoas_data=qdoas_data[:,[1,3,2,0]];
                            elif sensor_name=="SCIAMACHY":
                               qdoas_data=qdoas_data[:,[0,2,3,1]];

                         if sensor_name=="GOME" and varname=="Pixel type":
                            scan_direction=copy(qdoas_data)
                            scan_direction=scan_direction.astype(dtype=qdoas_field["type"]);
                            scan_direction[[x for (x,item) in enumerate(qdoas_data) if item==0 or item==1 or item==2]]=0;   # forward
                            scan_direction[where(qdoas_data==3)]=1;                                     # backscan
                            hp.AddVariable("scan_direction_type",
                                        data=scan_direction,
                                        description="scan direction for each measurement; enumeration values: forward (0), backward (1)",
                                        enum=["forward","backward"]);

                         qdoas_data=qdoas_data.astype(dtype=qdoas_field["type"])+qdoas_field["correction"];

                         hp.AddVariable(qdoas_field["harp_name"],
                                        data=qdoas_data,
                                        valid_min=qdoas_field["valid_min"],
                                        valid_max=qdoas_field["valid_max"],
                                        unit=qdoas_field["unit"],
                                        description=qdoas_field["description"]);

                   # Slant column densities

                   for species in species_dict:

                      if species not in HARP_species_dict:
                         print('['+functionName+'] '+species+' not recognized in HARP convention, ignored by this module');
                      else:

                         group_molec=species_dict[species];
                         group=group_molec[0];
                         molec=group_molec[1];

                         SlCol=sprintf('SlCol(%s)',molec);
                         SlErr=sprintf('SlErr(%s)',molec);

                         if group not in qdoas_group.groups:
                            print('['+functionName+'] '+'File ' + qdoasFile + ' do not contain ' + group + ' analysis window');
                         else:

                            if SlCol not in qdoas_group.groups[group].variables:
                               print('['+functionName+'] '+'File ' + qdoasFile + ' do not contain ' + molec + ' cross section in ' + group + ' analysis window');
                            elif hp.AddVariable(species+'_slant_column_number_density',
                                              data=asarray(qdoas_group.groups[group].variables[SlCol]).reshape(hdim),
                                              unit='molec/cm2',
                                              valid_min=-1.e300,
                                              valid_max=1.e300,
                                              fill_value=qdoas_group.groups[group].variables[SlCol]._FillValue,
                                              description=molec+' slant column number density retrieved from '+group+' analysis window')!=0:
                               print('['+functionName+'] '+'Can not create variable '+species+'_slant_column_number_density');

                            if SlErr in qdoas_group.groups[group].variables and \
                               hp.AddVariable(species+'_slant_column_number_density_uncertainty',
                                              data=asarray(qdoas_group.groups[group].variables[SlErr]).reshape(hdim),
                                              unit='molec/cm2',
                                              valid_min=-1.e300,
                                              valid_max=1.e300,
                                              fill_value=qdoas_group.groups[group].variables[SlErr]._FillValue,
                                              description='Uncertainty of the '+molec+' slant column number density retrieved from '+group+' analysis window')!=0:
                               print('['+functionName+'] '+'Can not create variable '+species+'_slant_column_number_density_uncertainty');



                   if (len(hp.harp_product)>0):
                      harp.export_product(hp.harp_product,harpFile,file_format="netcdf");
                   print(hp);

             nc.close();

             # Additional attributes

             if os.path.isfile(harpFile):

                nc=netCDF4.Dataset(harpFile,'r+',format='NETCDF4');
                nc.qdoas_file=os.path.basename(qdoasFile);
                if len(sensor_name)>0 and (sensor_name in QDOAS_satellite_sensors or sensor_name in QDOAS_groundbased_sensors):
                   nc.sensor_name=sensor_name;
                else:
                   nc.sensor_name="unknown"
                nc.close();


          except OSError:

             print('['+functionName+'] '+'Error while converting '+qdoasFile);

       return hp.harp_product;

